import platform
import re
import uuid
from fnmatch import fnmatch
from os import PathLike
from pathlib import Path
from typing import Union

import appdirs
import rarfile

from sickchill import settings
from sickchill.init_helpers import get_current_version

INSTANCE_ID = str(uuid.uuid1())
USER_AGENT = "SickChill/{version} ({os} {architecture} {os_version}; {instance})".format(
    version=get_current_version(), os=platform.system(), architecture=platform.machine(), os_version=platform.release(), instance=INSTANCE_ID
)

dateFormat = "%Y-%m-%d"
dateTimeFormat = "%Y-%m-%d %H:%M:%S"

# Mapping HTTP status codes to official W3C names
HTTP_STATUS_CODES = {
    300: "Multiple Choices",
    301: "Moved Permanently",
    302: "Found",
    303: "See Other",
    304: "Not Modified",
    305: "Use Proxy",
    306: "Switch Proxy",
    307: "Temporary Redirect",
    308: "Permanent Redirect",
    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Request Entity Too Large",
    414: "Request-URI Too Long",
    415: "Unsupported Media Type",
    416: "Requested Range Not Satisfiable",
    417: "Expectation Failed",
    418: "Im a teapot",
    419: "Authentication Timeout",
    420: "Enhance Your Calm",
    422: "Unprocessable Entity",
    423: "Locked",
    424: "Failed Dependency",
    426: "Upgrade Required",
    428: "Precondition Required",
    429: "Too Many Requests",
    431: "Request Header Fields Too Large",
    440: "Login Timeout",
    444: "No Response",
    449: "Retry With",
    450: "Blocked by Windows Parental Controls",
    451: [
        "Redirect",
        "Unavailable For Legal Reasons",
    ],
    494: "Request Header Too Large",
    495: "Cert Error",
    496: "No Cert",
    497: "HTTP to HTTPS",
    498: "Token expired/invalid",
    499: [
        "Client Closed Request",
        "Token required",
    ],
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Timeout",
    505: "HTTP Version Not Supported",
    506: "Variant Also Negotiates",
    507: "Insufficient Storage",
    508: "Loop Detected",
    509: "Bandwidth Limit Exceeded",
    510: "Not Extended",
    511: "Network Authentication Required",
    520: "CloudFlare - Web server is returning an unknown error",
    521: "CloudFlare - Web server is down",
    522: "CloudFlare - Connection timed out",
    523: "CloudFlare - Origin is unreachable",
    524: "CloudFlare - A timeout occurred",
    525: "CloudFlare - SSL handshake failed",
    526: "CloudFlare - Invalid SSL certificate",
    598: "Network read timeout error",
    599: "Network connect timeout error",
}
MEDIA_EXTENSIONS = [
    "3gp",
    "avi",
    "divx",
    "dvr-ms",
    "f4v",
    "flv",
    "m2ts",
    "m4v",
    "mkv",
    "mov",
    "mp4",
    "mpeg",
    "mpg",
    "ogm",
    "ogv",
    "rmvb",
    "tp",
    "ts",
    "vob",
    "webm",
    "wmv",
    "wtv",
]

SUBTITLE_EXTENSIONS = ["ass", "idx", "srt", "ssa", "sub"]
timeFormat = "%A %I:%M %p"


def http_code_description(http_code):
    """
    Get the description of the provided HTTP status code.
    :param http_code: The HTTP status code
    :return: The description of the provided ``http_code``
    """

    description = HTTP_STATUS_CODES.get(try_int(http_code))

    if isinstance(description, list):
        return f"({', '.join(description)})"

    return description


def get_extension(path: Union[Path, PathLike, str] = None, lower: bool = False) -> str:
    path = Path(path)
    result = path.suffix.lstrip(".")
    if lower:
        result = result.lower()

    return result


def is_sync_file(filename: Union[Path, PathLike, str] = None) -> bool:
    """
    Check if the provided ``filename`` is a sync file, based on its name.
    :param filename: The filename to check
    :return: ``True`` if the ``filename`` is a sync file, ``False`` otherwise
    """
    sync_extensions = settings.SYNC_FILES.split(",")
    return (
        filename.startswith(".syncthing")
        or get_extension(filename, lower=True) in sync_extensions
        or any(fnmatch(filename, match) for match in sync_extensions)
    )


def is_torrent_or_nzb_file(filename: Union[Path, PathLike, str] = None) -> bool:
    """
    Check if the provided ``filename`` is a NZB file or a torrent file, based on its extension.
    :param filename: The filename to check
    :return: ``True`` if the ``filename`` is a NZB file or a torrent file, ``False`` otherwise
    """
    return get_extension(filename, lower=True) in ("nzb", "torrent")


def is_media_file(filename):
    """
    Check if named file may contain media

    Parameters:
        filename: Filename to check
    Returns:
        True if this is a known media file, False if not
    """

    # ignore samples
    is_rar = is_rar_file(filename)

    path = Path(filename)
    if re.search(r"(^|[\W_])(?<!shomin.)(sample\d*)[\W_]", path.name, re.I):
        return False

    # ignore RARBG release intro
    if re.search(r"^RARBG\.(\w+\.)?(mp4|avi|txt)$", path.name, re.I):
        return False

    # ignore Kodi tvshow trailers
    if path.name == "tvshow-trailer.mp4":
        return False

    # ignore MACOS's retarded "resource fork" files
    if path.name.startswith("._"):
        return False

    if re.search("extras?$", path.name, re.I):
        return False

    return (get_extension(path, lower=True) in MEDIA_EXTENSIONS) or (is_rar and settings.UNPACK == settings.UNPACK_PROCESS_INTACT)


def is_rar_file(filename: Union[Path, PathLike, str]) -> bool:
    """
    Check if file is a RAR file, or part of a RAR set

    Parameters:
        filename: Filename to check
    Returns:
         True if this is RAR/Part file, False if not
    """
    result = False
    path = Path(filename)
    archive_regex = r"(?P<file>^(?P<base>(?:(?!\.part\d+\.rar$).)*)\.(?:(?:part0*1\.)?rar)$)"
    try:
        if path.is_file():
            result = rarfile.is_rarfile(path)
        else:
            result = re.search(archive_regex, path.name) is not None
    except (IOError, OSError):
        return result

    return result


def pretty_file_size(size, use_decimal=False, **kwargs):
    """
    Return a human-readable representation of the provided ``size``.

    :param size: The size to convert
    :param use_decimal: use decimal instead of binary prefixes (e.g. kilo = 1000 instead of 1024)

    :keyword units: A list of unit names in ascending order.
        Default units: ['B', 'KB', 'MB', 'GB', 'TB', 'PB']

    :return: The converted size
    """
    try:
        size = max(float(size), 0.0)
    except (ValueError, TypeError):
        size = 0.0

    remaining_size = size
    units = kwargs.pop("units", ["B", "KB", "MB", "GB", "TB", "PB"])
    block = 1024.0 if not use_decimal else 1000.0
    for unit in units:
        if remaining_size < block:
            return "{0:3.2f} {1}".format(remaining_size, unit)
        remaining_size /= block
    return size


def convert_size(size, default=None, use_decimal=False, **kwargs):
    """
    Convert a file size into the number of bytes

    :param size: to be converted
    :param default: value to return if conversion fails
    :param use_decimal: use decimal instead of binary prefixes (e.g. kilo = 1000 instead of 1024)

    :keyword sep: Separator between size and units, default is space
    :keyword units: A list of (uppercase) unit names in ascending order.
        Default units: ['B', 'KB', 'MB', 'GB', 'TB', 'PB']

    :keyword default_units: Default unit if none is given,
        default is the lowest unit on the scale, e.g. bytes

    :returns: the number of bytes, the default value, or 0
    """
    result = None

    try:
        sep = kwargs.pop("sep", " ")
        scale = kwargs.pop("units", ["B", "KB", "MB", "GB", "TB", "PB"])
        default_units = kwargs.pop("default_units", scale[0])

        if sep:
            size_tuple = size.strip().split(sep)
            scalar, units = size_tuple[0], size_tuple[1:]
            units = units[0].upper() if units else default_units
        else:
            regex_scalar = re.search(r"([\d. ]+)", size, re.I)
            scalar = regex_scalar.group() if regex_scalar else -1
            units = size.strip(scalar) if scalar != -1 else "B"

        scalar = float(scalar)
        scalar *= (1024 if not use_decimal else 1000) ** scale.index(units)

        result = scalar

    # TODO: Make sure fallback methods obey default units
    except AttributeError:
        result = size if size is not None else default

    except ValueError:
        result = default

    finally:
        try:
            if result != default:
                result = max(int(result), 0)
        except (TypeError, ValueError):
            pass

    return result


def remove_extension(filename: Union[Path, PathLike, str]) -> Union[Path, PathLike, str]:
    """
    Remove the extension of the provided ``filename``.
    The extension is only removed if it is in MEDIA_EXTENSIONS or ['nzb', 'torrent'].
    :param filename: The filename from which we want to remove the extension
    :return: The ``filename`` without its extension.
    """
    path = Path(filename)
    if not path.name:
        return filename

    is_media = get_extension(path, lower=True) in ["nzb", "torrent"] + MEDIA_EXTENSIONS
    return type(filename)((path, path.with_suffix(""))[is_media])


def replace_extension(filename: Union[Path, PathLike, str], new_extension: str) -> Union[Path, PathLike, str]:
    """
    Replace the extension of the provided ``filename`` with a new extension.
    :param filename: The filename for which we want to change the extension
    :param new_extension: The new extension to apply on the ``filename``
    :return: The ``filename`` with the new extension
    """
    if not isinstance(new_extension, (PathLike, str)):
        raise TypeError()

    path = Path(filename)
    if not path.name:
        return filename
    if not path.suffix:
        return filename
    else:
        if new_extension and not new_extension.startswith("."):
            new_extension = f".{new_extension}"

    return type(filename)(path.with_suffix(new_extension))


def sanitize_filename(filename):
    """
    Remove specific characters from the provided ``filename``.
    :param filename: The filename to clean
    :return: The ``filename``cleaned
    """

    if isinstance(filename, bytes):
        filename = filename.decode()

    if isinstance(filename, str):
        filename = re.sub(r"[\\/*]", "-", filename)
        filename = re.sub(r'[:"<>|?]', "", filename)
        filename = re.sub(r"™|-u2122", "", filename)  # Trademark Sign unicode: \u2122
        filename = filename.strip(" .")

        return filename

    return ""


def try_int(candidate, default_value=0):
    """
    Try to convert ``candidate`` to int, or return the ``default_value``.
    :param candidate: The value to convert to int
    :param default_value: The value to return if the conversion fails
    :return: ``candidate`` as int, or ``default_value`` if the conversion fails
    """

    try:
        return int(candidate)
    except (ValueError, TypeError):
        return default_value


def try_float(candidate, default_value=0):
    """
    Try to convert ``candidate`` to int, or return the ``default_value``.
    :param candidate: The value to convert to int
    :param default_value: The value to return if the conversion fails
    :return: ``candidate`` as int, or ``default_value`` if the conversion fails
    """

    try:
        return float(candidate)
    except (ValueError, TypeError):
        return default_value


def episode_num(season=None, episode=None, **kwargs):
    """
    Convert season and episode into string

    :param season: Season number
    :param episode: Episode Number
    :keyword numbering: Absolute for absolute numbering
    :returns: a string in s01e01 format or absolute numbering
    """

    numbering = kwargs.pop("numbering", "standard")

    if season or episode:
        if numbering == "standard":
            if season is not None and episode is not None:
                return f"S{season:02}E{episode:02}"
        elif numbering == "absolute":
            if episode is None:
                return f"{season:03}"


def choose_data_dir(program_dir) -> Path:
    old_data_dir = Path(program_dir).parent
    old_profile_path = Path.home().joinpath("sickchill")
    proper_data_dir = Path(appdirs.user_config_dir(appname="sickchill"))
    for location in [old_data_dir, old_profile_path, proper_data_dir]:
        for check in ["sickbeard.db", "sickchill.db", "config.ini"]:
            if location.joinpath(check).exists():
                return location.resolve()
    return proper_data_dir.resolve()



from pathlib import Path

from . import aniDBfileInfo as fileInfo


class TvDBMap(object):
    def __init__(self, cache_dir: Path):
        self.xmlMap = fileInfo.read_tvdb_map_xml(cache_dir)

    def get_tvdb_for_anidb(self, anidb_id):
        return self._get_x_for_y(anidb_id, "anidbid", "tvdbid")

    def get_anidb_for_tvdb(self, tvdb_id):
        return self._get_x_for_y(tvdb_id, "tvdbid", "anidbid")

    def _get_x_for_y(self, xValue, x, y):
        # print("searching "+x+" with the value "+str(xValue)+" and want to give back "+y)
        xValue = str(xValue)
        for anime in self.xmlMap.findall("anime"):
            try:
                if anime.get(x, False) == xValue:
                    return int(anime.get(y, 0))
            except ValueError as error:
                continue
        return 0

    def get_season_episode_for_anidb_absolute_number(self, anidb_id, absoluteNumber):
        # NOTE: this cant be done without the length of each season from thetvdb
        # TODO: implement
        season = 0
        episode = 0

        for anime in self.xmlMap.findall("anime"):
            if int(anime.get("anidbid", False)) == anidb_id:
                defaultSeason = int(anime.get("defaulttvdbseason", 1))

        return season, episode

    @staticmethod
    def get_season_episode_for_tvdb_absoluteNumber(anidb_id, absoluteNumber):
        # TODO: implement
        season = 0
        episode = 0
        return season, episode



from flask import Blueprint, render_template

from sickchill import logger, settings

# from sickchill.movies import movie

blueprint = Blueprint("movies", __name__, template_folder="templates", static_folder="static", url_prefix="/movies")


@blueprint.route("/")
def movies():
    logger.info("Loading movies page")
    logger.debug(f"movies: {settings.show_list}")

    return render_template("movies.html", movies=settings.movie_list)



import json
from pathlib import Path

from sickchill.oldbeard import config
from sickchill.oldbeard.browser import folders_at_path

from .. import settings
from .index import WebRoot
from .routes import Route


@Route("/browser(/?.*)", name="filebrowser")
class WebFileBrowser(WebRoot):
    def index(self):
        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        self.set_header("Content-Type", "application/json")

        path = self.get_argument("path")
        include_files = config.checkbox_to_value(self.get_argument("includeFiles", "false"))
        file_types = self.get_argument("fileTypes", "*").split(",")

        path = Path(path)
        if not path.exists():
            path = Path(settings.DATA_DIR)
        return json.dumps(folders_at_path(path, include_parent=True, include_files=include_files, file_types=file_types))

    def complete(self):
        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        self.set_header("Content-Type", "application/json")

        include_files = config.checkbox_to_value(self.get_argument("includeFiles", "false"))
        term = self.get_argument("term")
        file_types = self.get_argument("fileTypes", "*").split(",")

        paths = []
        for entry in folders_at_path(Path(term), include_parent=True, include_files=include_files, file_types=file_types):
            if "path" in entry:
                paths.append(entry.get("path", ""))

        return json.dumps(paths)



#!/usr/bin/env python3

import datetime
import mimetypes
import os
import platform
import shutil
import signal
import subprocess
import sys
import threading
import time
import traceback
import zipfile
from operator import attrgetter
from pathlib import Path
from typing import List, Union

import sickchill.start

try:
    from frontend.app import FlaskServer
except (ModuleNotFoundError, ImportError):
    FlaskServer = None

from sickchill import logger, settings
from sickchill.helper.common import choose_data_dir
from sickchill.init_helpers import check_installed, get_current_version, remove_pid_file, setup_gettext
from sickchill.movies import MovieList
from sickchill.oldbeard.name_parser.parser import NameParser, ParseResult

setup_gettext()

mimetypes.add_type("text/css", ".css")
# noinspection SpellCheckingInspection
mimetypes.add_type("application/sfont", ".otf")
# noinspection SpellCheckingInspection
mimetypes.add_type("application/sfont", ".ttf")
mimetypes.add_type("application/javascript", ".js")
mimetypes.add_type("application/font-woff", ".woff")

from configobj import ConfigObj

from sickchill.helper.argument_parser import SickChillArgumentParser
from sickchill.oldbeard import db, name_cache, network_timezones
from sickchill.oldbeard.event_queue import Events
from sickchill.tv import TVShow
from sickchill.update_manager import PipUpdateManager, UpdateManager
from sickchill.views.server_settings import SCWebServer

# http://bugs.python.org/issue7980#msg221094
THROWAWAY = datetime.datetime.strptime("20110101", "%Y%m%d")

signal.signal(signal.SIGINT, sickchill.start.sig_handler)
signal.signal(signal.SIGTERM, sickchill.start.sig_handler)


class SickChill:
    """
    Main SickChill module
    """

    def __init__(self):
        # system event callback for shutdown/restart
        settings.events = Events(self.shutdown)

        # daemon constants
        self.run_as_daemon = False

        # web server constants
        self.flask_server = None

        self.web_server = None
        self.forced_port = None
        self.no_launch = False

        self.start_port = settings.WEB_PORT

        self.console_logging = True

    @staticmethod
    def clear_cache():
        """
        Remove the Mako cache directory
        """
        # noinspection PyBroadException
        try:
            cache_folder = os.path.join(settings.CACHE_DIR, "mako")
            if os.path.isdir(cache_folder):
                shutil.rmtree(cache_folder)
        except Exception:
            logger.warning("Unable to remove the cache/mako directory!")

    def start(self):
        """
        Start SickChill
        """
        # do some preliminary stuff
        settings.MY_FULLNAME = os.path.normpath(os.path.abspath(__file__))
        settings.MY_NAME = os.path.basename(settings.MY_FULLNAME)

        settings.DATA_DIR = choose_data_dir(settings.PROG_DIR)

        settings.MY_ARGS = sys.argv[1:]

        # Rename the main thread
        threading.current_thread().name = "MAIN"

        args = SickChillArgumentParser(settings.DATA_DIR).parse_args()

        # Add methods here when you want to perform an action and exit, without starting the webserver
        if args.subparser_name == "test-name" and args.name:
            results = []
            for parser in args.parser.lower().split(","):
                result = self.test_name(args.name, parser if parser != "all" else "")
                results.append(result)
                parser_name = parser or "all"
                self.log(f"{parser_name}: {result}")
            sys.exit(int(not any(results)))

        if args.no_update:
            settings.DISABLE_UPDATER = True
        elif args.force_update:
            result = self.force_update()
            sys.exit(int(not result))  # Ok -> 0 , Error -> 1
        elif args.install_file:
            result = self.install_file(Path(args.install_file))
            sys.exit(int(not result))  # Ok -> 0 , Error -> 1

        settings.NO_RESIZE = args.noresize
        self.console_logging = not (hasattr(sys, "frozen") or args.quiet or args.daemon)
        self.no_launch = args.nolaunch or args.daemon or args.debug
        self.forced_port = args.port
        self.run_as_daemon = args.daemon and platform.system() != "Windows"

        # The pid file is only useful in daemon mode, make sure we can write the file properly
        if bool(args.pidfile) and not self.run_as_daemon:
            self.log("Not running in daemon mode. PID file creation disabled")

        settings.DATA_DIR = Path(args.datadir).resolve() if args.datadir else settings.DATA_DIR
        settings.CONFIG_FILE = str(Path(args.config).resolve() if args.config else settings.DATA_DIR.joinpath("config.ini"))

        # Make sure that we can create the data dir
        if not os.access(settings.DATA_DIR, os.F_OK):
            try:
                settings.DATA_DIR.mkdir(mode=0x0744, parents=True, exist_ok=True)
            except os.error:
                raise SystemExit("Unable to create data directory: {}".format(settings.DATA_DIR))

        # Make sure we can write to the data dir
        if not os.access(settings.DATA_DIR, os.W_OK):
            raise SystemExit("Data directory must be writeable: {}".format(settings.DATA_DIR))

        # Make sure we can write to the config file
        if not os.access(settings.CONFIG_FILE, os.W_OK):
            if os.path.isfile(settings.CONFIG_FILE):
                raise SystemExit("Config file must be writeable: {}".format(settings.CONFIG_FILE))
            elif not os.access(os.path.dirname(settings.CONFIG_FILE), os.W_OK):
                raise SystemExit("Config file root dir must be writeable: {}".format(os.path.dirname(settings.CONFIG_FILE)))

        os.chdir(settings.DATA_DIR)

        # Check if we need to perform a restore first
        restore_dir = os.path.join(settings.DATA_DIR, "restore")
        if os.path.exists(restore_dir):
            self.log(f"Found restore directory {restore_dir}, restoring your backup!")
            success = self.restore_db(restore_dir, settings.DATA_DIR)
            status = ("FAILED", "SUCCESSFUL")[success]
            self.log(f"Restore: restoring DB and config.ini {status}!", not success)

        upgrade_dir = Path(settings.DATA_DIR) / "force-install"
        if upgrade_dir.is_dir():
            upgrade_files = []
            for upgrade_file in upgrade_dir.glob("sickchill-*-none-any.whl"):
                if upgrade_file.is_file():
                    upgrade_files.append(upgrade_file)

            for upgrade_file in upgrade_dir.glob("Result.zip"):
                if upgrade_file.is_file():
                    upgrade_files.append(upgrade_file)

            if upgrade_files:
                # The only files in here are a wheel and/or a Result.zip
                if len(list(upgrade_dir.iterdir())) == len(upgrade_files):
                    self.log(f"Found force-install directory {upgrade_dir} with {upgrade_files[0].name}, processing")
                    result = self.install_file(upgrade_files[0])
                    self.log(f"Removing sickchill*.whl and Results.zip from force-install directory {upgrade_dir}")
                    for file in upgrade_files:
                        file.unlink()

                    sys.exit(int(not result))  # Ok -> 0 , Error -> 1
                else:
                    self.log(f"Cannot process this upgrade directory, you have files that don't belong in it. Continuing with startup", 1)
                    sys.exit(1)  # Ok -> 0 , Error -> 1

        # Load the config and publish it to the oldbeard package
        if not os.path.isfile(settings.CONFIG_FILE):
            self.log(f"Unable to find {settings.CONFIG_FILE}, all settings will be default!")

        settings.CFG = ConfigObj(settings.CONFIG_FILE, encoding="UTF-8", indent_type="  ")

        # Initialize the config and our threads
        sickchill.start.initialize(
            console_logging=self.console_logging, debug=args.debug, dbdebug=args.dbdebug, disable_file_logging=args.no_file_logging or args.debug
        )

        # Get PID
        settings.PID = os.getpid()

        # Build from the DB to start with
        self.load_shows_from_db()

        logger.info("Starting SickChill [{version}] using '{config}'".format(version=get_current_version(), config=settings.CONFIG_FILE))

        self.clear_cache()

        if settings.DEVELOPER:
            settings.movie_list = MovieList()

        web_options = {"debug": args.debug}
        if self.forced_port:
            logger.info("Forcing web server to port {port}".format(port=self.forced_port))
            self.start_port = self.forced_port
            web_options.update(
                {
                    "port": int(self.start_port),
                }
            )
        else:
            self.start_port = settings.WEB_PORT

        # start web server
        self.web_server = SCWebServer(web_options)
        self.web_server.start()

        if args.flask and FlaskServer:
            # start the flask frontend
            if args.flask_port:
                port = args.flask_port
            else:
                port = int(self.start_port) + 1

            if args.flask_host:
                web_host = args.flask_host
            else:
                if settings.WEB_HOST and settings.WEB_HOST != "0.0.0.0":
                    web_host = settings.WEB_HOST
                else:
                    web_host = ("0.0.0.0", "")[settings.WEB_IPV6]

            self.flask_server = FlaskServer(web_host, port)
            self.flask_server.start()

        # Fire up all our threads
        sickchill.start.start()

        # Build internal name cache
        name_cache.build_name_cache()

        # Pre-populate network timezones, it isn't thread safe
        network_timezones.update_network_dict()

        # Check for metadata indexer updates for shows (sets the next aired ep!)
        # oldbeard.showUpdateScheduler.forceRun()

        # Launch browser
        if settings.LAUNCH_BROWSER and not self.no_launch:
            sickchill.start.launchBrowser("https" if settings.ENABLE_HTTPS else "http", self.start_port, settings.WEB_ROOT)

        # main loop
        while True:
            time.sleep(1)

    @staticmethod
    def load_shows_from_db():
        """
        Populates the show list with shows from the database
        """
        logger.debug("Loading initial show list")

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT indexer, indexer_id, location FROM tv_shows;")

        settings.show_list = []
        for sql_show in sql_results:
            if settings.stopping or settings.restarting:
                break

            try:
                cur_show = TVShow(sql_show["indexer"], sql_show["indexer_id"])
                cur_show.next_episode()
                settings.show_list.append(cur_show)
            except Exception as error:
                logger.exception("There was an error creating the show in {}: Error {}".format(sql_show["location"], error))
                logger.debug(traceback.format_exc())
            except KeyboardInterrupt:
                break

        # Presort show_list, so we don't have to do it every page load
        settings.show_list = sorted(settings.show_list, key=attrgetter("sort_name"))

    @staticmethod
    def restore_db(src_dir, dst_dir):
        """
        Restore the Database from a backup

        :param src_dir: Directory containing backup
        :param dst_dir: Directory to restore to
        :return:
        """
        # noinspection PyBroadException
        try:
            files_list = ["sickbeard.db", "sickchill.db", "config.ini", "failed.db", "cache.db"]
            for filename in files_list:
                src_file = os.path.join(src_dir, filename)
                dst_file = os.path.join(dst_dir, filename)
                bak_file = os.path.join(dst_dir, "{}.bak-{}".format(filename, datetime.datetime.now().strftime("%Y%m%d_%H%M%S")))
                sickchill_db = os.path.join(dst_dir, "sickchill.db")
                sickbeard_db = os.path.join(src_dir, "sickbeard.db")
                if os.path.isfile(src_file):
                    if src_file == sickbeard_db:
                        dst_file = sickchill_db

                    if os.path.isfile(dst_file):
                        shutil.move(dst_file, bak_file)
                    shutil.move(src_file, dst_file)

            sickbeard_db = os.path.join(dst_dir, "sickbeard.db")
            sickchill_db = os.path.join(dst_dir, "sickchill.db")
            if os.path.isfile(sickbeard_db) and not os.path.isfile(sickchill_db):
                shutil.move(sickbeard_db, sickchill_db)

            return True
        except Exception:
            return False

    def shutdown(self, event):
        """
        Shut down SickChill

        :param event: Type of shutdown event, used to see if restart required
        """
        if settings.started:
            sickchill.start.halt()  # stop all tasks
            sickchill.start.save_all()  # save all shows to DB

            # shutdown web server
            if self.web_server:
                logger.info("Shutting down Tornado")
                self.web_server.shutdown()

                # noinspection PyBroadException
                try:
                    self.web_server.join(10)
                except Exception:
                    pass

            self.clear_cache()  # Clean cache

            # if run as daemon delete the pid file
            remove_pid_file()

            if event == sickchill.oldbeard.event_queue.Events.SystemEvent.RESTART:
                popen_list = [sys.executable, settings.MY_FULLNAME]
                if popen_list and not settings.NO_RESTART:
                    popen_list += settings.MY_ARGS
                    # noinspection SpellCheckingInspection
                    if "--nolaunch" not in popen_list:
                        # noinspection SpellCheckingInspection
                        popen_list += ["--nolaunch"]
                    logger.info("Restarting SickChill with {options}".format(options=popen_list))
                    # shutdown the logger to make sure it's released the logfile BEFORE it restarts SC.
                    logger.shutdown()
                    subprocess.Popen(popen_list, cwd=os.getcwd(), universal_newlines=True)

        # Make sure the logger has stopped, just in case
        logger.shutdown()
        os._exit(0)  # noqa

    def force_update(self):
        """
        Forces SickChill to update to the latest version and exit.

        :return: True if successful, False otherwise
        """
        if not check_installed():
            self.log("Sickchill updater no longer works with git or source installs", 1)
            return False

        self.log("Forcing SickChill to update using pip...")
        if not PipUpdateManager().update():
            self.log("Failed to force an update", 1)
            return False

        self.log("Successfully updated to the latest pip release. You may now run SickChill normally")
        return True

    def install_file(self, file: Path):
        """
        Installs a wheel file, potentially inside a zip file
        :return: True if successful, False otherwise
        """

        if not check_installed():
            self.log("Sickchill updater no longer works with git or source installs", 1)
            return False

        file = file.resolve()

        if not file.is_file():
            self.log(f"File to install was not found ({file})", 1)
            return False

        if file.name.endswith("zip"):
            self.log("File passed with to install was a zip file, extracting the wheel from it")
            if not zipfile.is_zipfile(file):
                self.log("File passed to install is not a valid zip file", 1)
                return False

            try:
                with zipfile.ZipFile(file) as zf:
                    for filename in zf.namelist():
                        if filename.endswith("whl"):
                            extracted = zf.extract(filename, path=file.parent)
                            file = Path(extracted).resolve()
                            break
            except zipfile.BadZipfile as error:
                self.log(f"Error while extracting {file}: {error}", 1)
                return False

        updater = UpdateManager()
        if not updater.updater:
            self.log(f"Unable to install files, the updater is disabled", 1)
            return False

        self.log(f"Creating a backup of your database and config before installing {file.name}")
        if updater.backup():
            self.log(f"Installing the wheel {file.name}")
            if not updater.updater.pip_install(file):
                self.log(f"Failed to install {file.name}", 1)
                return False
        else:
            self.log("Could not make a backup, not installing the file", 1)
            return False

        self.log("Successfully installed to the wheel. You may now run SickChill normally")
        return True

    @staticmethod
    def test_name(name: str, parse_method="") -> List[Union[ParseResult, None]]:
        return NameParser(parse_method=parse_method)._parse_string(name)

    def log(self, message: str, error: Union[bool, int] = 0) -> None:
        if self.console_logging:
            if error > 0:
                sys.stderr.write(f"{message}\n")
            else:
                sys.stdout.write(f"{message}\n")


def main():
    # start SickChill
    SickChill().start()
    remove_pid_file()


if __name__ == "__main__":
    main()



import logging
import logging.handlers
import os
import re
import sys
from logging import NullHandler
from urllib.parse import quote

from sickchill import settings
from sickchill.helper.common import dateTimeFormat
from sickchill.oldbeard import classes, notifiers

# log levels
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG
DB = 5

LOGGING_LEVELS = {
    "ERROR": ERROR,
    "WARNING": WARNING,
    "INFO": INFO,
    "DEBUG": DEBUG,
    "DB": DB,
}

censored_items = {}


class DispatchFormatter(logging.Formatter, object):
    """
    Censor information such as API keys, usernames, and passwords from the Log
    """

    def __init__(self, fmt=None, datefmt=None, style="{"):
        super().__init__(fmt, datefmt, style=style)

    def format(self, record):
        """
        Strips censored items from string and formats the log line

        :param record: to format
        """

        msg = record.msg

        # set of censored items
        censored = {item for _, item in censored_items.items() if item}
        # set of censored items and urlencoded counterparts
        censored = list(censored | {quote(item) for item in censored})
        # sort the list in order of descending length so that entire item is censored
        # e.g. password and password_1 both get censored instead of getting ********_1
        censored.sort(key=len, reverse=True)

        if not isinstance(msg, (str, bytes)):
            msg = repr(msg)

        for item in censored:
            try:
                # passwords that include ++ for example will error. Cannot escape or it wont match at all.
                msg = re.sub(rf"\b({item})\b", "*" * 8, msg)
            except re.error:
                msg = msg.replace(item, "*" * 8)
            except TypeError:
                print(msg)

        # Needed because Newznab apikey isn't stored as key=value in a section.
        msg = re.sub(r"([&?]r|[&?]apikey|[&?]jackett_apikey|[&?]api_key)(?:=|%3D)[^&]*([&\w]?)", r"\1=**********\2", msg, re.I)

        if record.levelno == ERROR:
            classes.ErrorViewer.add(classes.UIError(msg))
            notifiers.notify_logged_error(classes.UIError(msg))

        elif record.levelno == WARNING:
            classes.WarningViewer.add(classes.UIError(msg))

        return super().format(record)


class Logger(object):
    """
    Logger to create log entries
    """

    def __init__(self):
        self.logger = logging.getLogger("sickchill")

        self.loggers = [
            logging.getLogger("sickchill"),
            logging.getLogger("sickchill.movie"),
            logging.getLogger("tornado.general"),
            logging.getLogger("tornado.application"),
            logging.getLogger("subliminal"),
            logging.getLogger("tornado.access"),
            logging.getLogger("imdbpy.parser.http.piculet"),
            logging.getLogger("imdbpy.parser.http.domparser"),
            logging.getLogger("sqlalchemy.engine"),
            logging.getLogger("sqlalchemy.pool"),
            logging.getLogger("sqlalchemy.dialect"),
            logging.getLogger("imdbpy"),
            logging.getLogger("imdbpy.parser.s3"),
        ]

        self.console_logging = False
        self.file_logging = False
        self.debug_logging = False
        self.database_logging = False
        self.log_file = None

    def init_logging(self, console_logging=False, file_logging=False, debug_logging=False, database_logging=False):
        """
        Initialize logging

        :param console_logging: True if logging to console
        :param file_logging: True if logging to file
        :param debug_logging: True if debug logging is enabled
        :param database_logging: True if logging database access
        """
        self.log_file = self.log_file or os.path.join(settings.LOG_DIR, "sickchill.log")

        global log_file
        log_file = self.log_file

        self.debug_logging = debug_logging
        self.console_logging = console_logging
        self.file_logging = file_logging
        self.database_logging = database_logging

        logging.addLevelName(DB, "DB")  # add a new logging level DB
        logging.getLogger().addHandler(NullHandler())  # nullify root logger

        # set custom root logger
        for logger in self.loggers:
            if logger is not self.logger:
                logger.root = self.logger
                logger.parent = self.logger

        log_level = DB if self.database_logging else DEBUG if self.debug_logging else INFO

        # set minimum logging level allowed for loggers
        for logger in self.loggers:
            if logger.name in ("subliminal", "tornado.access", "tornado.general", "imdbpy.parser.http.piculet", "imdbpy.parser.http.domparser"):
                logger.setLevel("CRITICAL")
            elif (logger.name.startswith("sqlalchemy") or logger.name.startswith("imdb")) and not self.database_logging:
                logger.setLevel("WARNING")
            else:
                logger.setLevel(log_level)

        log_format = "{asctime} {levelname} :: {threadName} :: {message}"
        # console log handler
        if self.console_logging:
            console = logging.StreamHandler()
            console.setFormatter(DispatchFormatter(log_format, dateTimeFormat))
            console.setLevel(log_level)

            for logger in self.loggers:
                logger.addHandler(console)

        # rotating log file handler
        if self.file_logging:
            rfh = logging.handlers.RotatingFileHandler(self.log_file, maxBytes=int(settings.LOG_SIZE * 1048576), backupCount=settings.LOG_NR)
            rfh.setFormatter(DispatchFormatter(log_format, dateTimeFormat))
            rfh.setLevel(log_level)

            for logger in self.loggers:
                logger.addHandler(rfh)

    def restart(self, change_log_dir: bool = False):
        """
        Sets the logging level for root and all of our loggers
        """
        if self.debug_logging == settings.DEBUG and self.database_logging == settings.DBDEBUG and not change_log_dir:
            return

        self.debug_logging = settings.DEBUG
        self.database_logging = settings.DBDEBUG

        if not change_log_dir:
            self.logger.info(f"Changing DEBUG to {settings.DEBUG} and DATABASE DEBUG to {settings.DBDEBUG}")

        # Set these back to None, so they are reset on init
        global log_file
        self.log_file = None
        log_file = self.log_file

        self.close_and_remove_handlers()
        self.init_logging(self.console_logging, self.file_logging, self.debug_logging, self.database_logging)

    def close_and_remove_handlers(self):
        for logger in self.loggers:
            for handler in logger.handlers[:]:
                handler.close()
                logger.removeHandler(handler)

    def shutdown(self):
        """
        Shut down the logger
        """
        self.close_and_remove_handlers()
        logging.shutdown()

    def log_error_and_exit(self, error_msg, *args, **kwargs):
        self.logger.log(ERROR, error_msg, *args, **kwargs)

        if not self.console_logging:
            sys.exit(error_msg)
        else:
            sys.exit(1)


class Wrapper(object):
    instance = Logger()

    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __getattr__(self, name):
        try:
            return getattr(self.wrapped, name)
        except AttributeError:
            return getattr(self.instance, name)


_globals = sys.modules[__name__] = Wrapper(sys.modules[__name__])

LOG_FILTERS = {
    "<NONE>": _("&lt;No Filter&gt;"),
    "DAILYSEARCHER": _("Daily Searcher"),
    "BACKLOG": _("Backlog"),
    "SHOWUPDATER": _("Show Updater"),
    "CHECKVERSION": _("Check Version"),
    "SHOWQUEUE": _("Show Queue"),
    "SEARCHQUEUE": _("Search Queue (All)"),
    "SEARCHQUEUE-DAILY-SEARCH": _("Search Queue (Daily Searcher)"),
    "SEARCHQUEUE-BACKLOG": _("Search Queue (Backlog)"),
    "SEARCHQUEUE-MANUAL": _("Search Queue (Manual)"),
    "SEARCHQUEUE-RETRY": _("Search Queue (Retry/Failed)"),
    "SEARCHQUEUE-RSS": _("Search Queue (RSS)"),
    "FINDPROPERS": _("Find Propers"),
    "POSTPROCESSOR": _("Postprocessor"),
    "FINDSUBTITLES": _("Find Subtitles"),
    "TRAKTCHECKER": _("Trakt Checker"),
    "EVENT": _("Event"),
    "ERROR": _("Error"),
    "TORNADO": _("Tornado"),
    "Thread": _("Thread"),
    "MAIN": _("Main"),
}


def log_data(min_level, log_filter, log_search, max_lines):
    regex = r"^\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}:\d{2} ([A-Z]+) \:\: (.+?) \:\: (.*)$"
    if log_filter not in LOG_FILTERS:
        log_filter = "<NONE>"

    final_data = []

    log_files = []
    if os.path.isfile(Wrapper.instance.log_file):
        log_files.append(Wrapper.instance.log_file)

        for i in range(1, int(settings.LOG_NR)):
            name = f"{Wrapper.instance.log_file}.{i}"
            if not os.path.isfile(name):
                break
            log_files.append(name)
    else:
        return final_data

    data = []
    for _log_file in log_files:
        if len(data) < max_lines:
            with open(_log_file, "r") as f:
                data += [line.strip() + "\n" for line in reversed(f.readlines()) if line.strip()]
        else:
            break

    found_lines = 0
    for x in data:
        match = re.match(regex, x)

        if match:
            level = match.group(1)
            log_name = match.group(2)

            if not settings.DEBUG and level == "DEBUG":
                continue

            if not settings.DBDEBUG and level == "DB":
                continue

            if level not in LOGGING_LEVELS:
                final_data.append("AA " + x)
                found_lines += 1
            elif log_search and log_search.lower() in x.lower():
                final_data.append(x)
                found_lines += 1
            elif not log_search and LOGGING_LEVELS[level] >= int(min_level) and (log_filter == "<NONE>" or log_name.startswith(log_filter)):
                final_data.append(x)
                found_lines += 1
        else:
            final_data.append("AA " + x)
            found_lines += 1

        if found_lines >= max_lines:
            break

    return final_data


log_error_and_exit = Wrapper.instance.log_error_and_exit
log_file = Wrapper.instance.log_file

log = Wrapper.instance.logger.log
debug = Wrapper.instance.logger.debug
info = Wrapper.instance.logger.info
warning = Wrapper.instance.logger.warning
error = Wrapper.instance.logger.error
exception = error
critical = Wrapper.instance.logger.critical


def database(msg, *args, **kwargs):
    """
    Handler to add database sql logging
    @param msg: message to be logged
    @param args: arguments to logger
    @param kwargs:  kwargs to logger
    """
    if settings.DBDEBUG:
        debug(msg, args, kwargs)


init_logging = Wrapper.instance.init_logging
restart = Wrapper.instance.restart
shutdown = Wrapper.instance.shutdown



from urllib.parse import urljoin

import requests
from requests import RequestException
from requests.structures import CaseInsensitiveDict

from sickchill import logger, settings


class Notifier(object):
    def _make_headers(self, emby_apikey=None):
        return CaseInsensitiveDict({"X-Emby-Token": emby_apikey or settings.EMBY_APIKEY, "Content-Type": "application/json"})

    def _notify_emby(self, message, host=None, emby_apikey=None):
        """Handles notifying Emby host via HTTP API

        Returns:
            Returns True for no issue or False if there was an error

        """
        url = urljoin(host or settings.EMBY_HOST, "emby/Notifications/Admin")
        params = {"Name": "SickChill", "Description": message, "ImageUrl": settings.LOGO_URL}

        if not settings.USE_EMBY:
            logger.debug("Notification for Emby not enabled, skipping this notification")
            return False

        try:
            response = requests.post(url, params=params, headers=self._make_headers(emby_apikey))
            if response:
                logger.debug(_("EMBY: HTTP response: {content}").format(content=response.content))
            response.raise_for_status()

            return True
        except RequestException as error:
            logger.warning(_("EMBY: Warning: Could not contact Emby at {url} {error}").format(url=url, error=error))
            return False

    ##############################################################################
    # Public functions
    ##############################################################################

    def test_notify(self, host, emby_apikey):
        return self._notify_emby(_("This is a test notification from SickChill"), host, emby_apikey)

    def update_library(self, show=None):
        """Handles updating the Emby Media Server host via HTTP API

        Returns:
            Returns True for no issue or False if there was an error

        """

        if settings.USE_EMBY:
            if not settings.EMBY_HOST:
                logger.debug(_("EMBY: No host specified, check your settings"))
                return False

            params = {}
            if show:
                params.update({"TvdbId": show.indexerid})
                # Endpoint emby/Library/Series/Added is deprecated http://swagger.emby.media/?staticview=true#/LibraryService/postLibrarySeriesAdded
                url = urljoin(settings.EMBY_HOST, "emby/Library/Media/Updated")
            else:
                url = urljoin(settings.EMBY_HOST, "emby/Library/Refresh")

            try:
                response = requests.post(url, params=params, headers=self._make_headers())
                response.raise_for_status()
                logger.debug(_("EMBY: HTTP status: {status_code}, response: {content}").format(status_code=response.status_code, content=response.content))
                return True

            except requests.exceptions.RequestException as error:
                logger.warning(_("EMBY: Warning: Could not contact Emby at {url} {error}").format(url=url, error=error))

                return False



import re
import urllib.parse
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import subliminal

from sickchill import logger, settings
from sickchill.helper.metaclasses import Singleton
from sickchill.oldbeard.classes import SearchResult

if TYPE_CHECKING:
    from sickchill.tv import TVEpisode, TVShow
    from sickchill.oldbeard.subtitles import Scores
    from sickchill.providers.GenericProvider import GenericProvider

from sickchill.helper.common import remove_extension, try_int
from sickchill.helper.exceptions import EpisodeNotFoundException
from sickchill.oldbeard.common import FAILED, Quality, SNATCHED, SUBTITLED, WANTED
from sickchill.oldbeard.db import DBConnection


class History(object, metaclass=Singleton):
    date_format = "%Y%m%d%H%M%S"

    def __init__(self):
        self.db: DBConnection = DBConnection()  # DataSource: sickchill.db
        self.failed_db: DBConnection = DBConnection("failed.db")  # DataSource: failed.db

        self.trim()

    def remove(self, items):
        """
        Removes the selected history
        :param items: Contains the properties of the log entries to remove
        """
        query = ""

        params = []
        for item in items:
            if query:
                query += " OR "

            query += "(date IN (?) AND showid = ? AND season = ? AND episode = ?)"
            params.extend([",".join(item["dates"]), item["show_id"], item["season"], item["episode"]])

        self.db.action("DELETE FROM history WHERE " + query, params)

    def clear(self):
        """
        Clear all the history
        """
        # noinspection SqlWithoutWhere
        self.db.action("DELETE FROM history")

    def get(self, limit: int = 100, action: str = None):
        """
        :param limit: The maximum number of elements to return
        :param action: The type of action to filter in the history. Either 'downloaded' or 'snatched'. Anything else or
                        no value will return everything (up to ``limit``)
        :return: The last ``limit`` elements of type ``action`` in the history
        """

        action = action.lower() if isinstance(action, str) else ""

        if action == "downloaded":
            actions = Quality.DOWNLOADED
        elif action == "snatched":
            actions = Quality.SNATCHED
        else:
            actions = Quality.DOWNLOADED + Quality.SNATCHED

        limit = max(try_int(limit, 0), 0)

        # DataSource: sickchill.db
        common_sql = (
            "SELECT action, date, episode, provider, h.quality, resource, season, show_name, showid "
            "FROM history h, tv_shows s "
            "WHERE h.showid = s.indexer_id "
        )  # DataSource: sickchill.db
        replacements = ",".join(["?"] * len(actions))
        filter_sql = f"AND action IN ({replacements})"
        order_sql = "ORDER BY date DESC "

        if limit == 0:
            if actions:
                results = self.db.select(common_sql + filter_sql + order_sql, actions)
            else:
                results = self.db.select(common_sql + order_sql)
        else:
            if actions:
                results = self.db.select(common_sql + filter_sql + order_sql + "LIMIT ?", actions + [limit])
            else:
                results = self.db.select(common_sql + order_sql + "LIMIT ?", [limit])

        data = []
        for result in results:
            data.append(
                {
                    "action": result["action"],
                    "date": result["date"],
                    "episode": result["episode"],
                    "provider": result["provider"],
                    "quality": result["quality"],
                    "resource": result["resource"],
                    "season": result["season"],
                    "show_id": result["showid"],
                    "show_name": result["show_name"],
                }
            )

        return data

    def trim(self):
        """
        Remove all elements older than 30 days from the history
        """
        back_thirty_days = (datetime.today() - timedelta(days=30)).strftime(self.date_format)
        self.db.action("DELETE FROM history WHERE date < ?", [back_thirty_days])
        if settings.USE_FAILED_DOWNLOADS:
            self.failed_db.action("DELETE FROM history WHERE date < ?", [back_thirty_days])

    def _log_history_item(self, action: int, showid: int, season: int, episode: int, quality: int, resource: str, provider, version=-1):
        """
        Insert a history item in DB

        :param action: action taken (snatch, download, etc)
        :param showid: showid this entry is about
        :param season: show season
        :param episode: show episode
        :param quality: media quality
        :param resource: resource used
        :param provider: provider used
        :param version: tracked version of file (defaults to -1)
        """
        # DataSource: sickchill.db
        return self.db.action(
            "INSERT INTO history (action, date, showid, season, episode, quality, resource, provider, version) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
            [action, datetime.today().strftime(self.date_format), showid, season, episode, quality, resource, provider, version],
        )

    def log_snatch(self, result: SearchResult):
        """
        Log history of snatch

        :param result: search result object
        """
        provider_class: "GenericProvider" = result.provider
        show: "TVShow" = result.episodes[0].show
        if provider_class:
            provider = provider_class.name
        else:
            provider = "unknown"

        for episode in result.episodes:
            self._log_history_item(
                Quality.compositeStatus(SNATCHED, result.quality),
                episode.show.indexerid,
                episode.season,
                episode.episode,
                result.quality,
                result.name,
                provider,
                result.version,
            )
            if settings.USE_FAILED_DOWNLOADS:
                self.failed_db.action(
                    'INSERT INTO history (date, size, "release", provider, showid, season, episode, old_status) VALUES (?, ?, ?, ?, ?, ?, ?, ?)',
                    [
                        datetime.today().strftime(self.date_format),
                        result.size,
                        self.prepare_failed_name(result.name),
                        provider,
                        show.indexerid,
                        episode.season,
                        episode.episode,
                        episode.status,
                    ],
                )

    def log_download(self, episode: "TVEpisode", filename: str, quality: int, group: str = None, version: int = -1):
        """
        Log history of download

        :param episode: episode of show
        :param filename: file on disk where the download is
        :param quality: Quality of download
        :param group: Release group
        :param version: Version of file (defaults to -1)
        """
        self._log_history_item(episode.status, episode.show.indexerid, episode.season, episode.episode, quality, filename, group or -1, version)

    def log_subtitle(self, show: int, season: int, episode: int, status: int, subtitle: subliminal.subtitle.Subtitle, scores: "Scores"):
        """
        Log download of subtitle

        :param show: Showid of download
        :param season: Show season
        :param episode: Show episode
        :param status: Status of download
        :param subtitle: Result object
        :param scores: Scores named tuple
        """
        if settings.SUBTITLES_HISTORY:
            logger.debug(
                f"[{subtitle.provider_name}] Subtitle score for {subtitle.id} is: {scores.res}/{scores.percent}% (min={scores.min}/{scores.min_percent})"
            )
            status, quality = Quality.splitCompositeStatus(status)
            self._log_history_item(
                Quality.compositeStatus(SUBTITLED, quality), show, season, episode, quality, subtitle.language.opensubtitles, subtitle.provider_name
            )

    def log_failed(self, episode_object: "TVEpisode", release: str, provider: str = ""):
        """
        Log a failed download

        :param episode_object: Episode object
        :param release: Release group
        :param provider: Provider used for snatch
        """

        status, quality = Quality.splitCompositeStatus(episode_object.status)
        self._log_history_item(
            Quality.compositeStatus(FAILED, quality), episode_object.show.indexerid, episode_object.season, episode_object.episode, quality, release, provider
        )

        if not settings.USE_FAILED_DOWNLOADS:
            return

        size = -1

        release = self.prepare_failed_name(release)

        sql_results = self.failed_db.select('SELECT * FROM history WHERE "release" = ?', [release])

        if not sql_results:
            logger.warning("Release not found in snatch history.")
        elif len(sql_results) > 1:
            logger.warning("Multiple logged snatches found for release")
            num_sizes = len(set(x["size"] for x in sql_results))
            providers = len(set(x["provider"] for x in sql_results))
            if num_sizes == 1:
                logger.warning("However, they're all the same size. Continuing with found size.")
                size = sql_results[0]["size"]
            else:
                logger.warning("They also vary in size. Deleting the logged snatches and recording this release with no size/provider")
                for result in sql_results:
                    self.remove_snatch(result["release"], result["size"], result["provider"])

            if providers == 1:
                logger.info("They're also from the same provider. Using it as well.")
                provider = sql_results[0]["provider"]
        else:
            size = sql_results[0]["size"]
            provider = sql_results[0]["provider"]

        if not self.has_failed(release, size, provider):
            self.failed_db.action('INSERT INTO failed ("release", size, provider) VALUES (?, ?, ?)', [release, size, provider])

        self.remove_snatch(release, size, provider)

    @staticmethod
    def prepare_failed_name(release: str):
        """Standardizes release name for failed DB"""

        fixed = urllib.parse.unquote(release)
        if fixed.endswith((".nzb", ".torrent")):
            fixed = remove_extension(fixed)

        return re.sub(r"[.\-+ ]", "_", fixed)

    def log_success(self, release):
        self.failed_db.action('DELETE FROM history WHERE "release" = ?', [self.prepare_failed_name(release)])

    def has_failed(self, release: str, size: int, provider: str = "%"):
        """
        Returns True if a release has previously failed.

        If provider is given, return True only if the release is found
        with that specific provider. Otherwise, return True if the release
        is found with any provider.

        :param release: Release name to record failure
        :param size: Size of release
        :param provider: Specific provider to search (defaults to all providers)
        :return: True if a release has previously failed.
        """

        return bool(
            self.failed_db.select_one(
                'SELECT "release" FROM failed WHERE "release" = ? AND size = ? AND provider LIKE ? ', [self.prepare_failed_name(release), size, provider]
            )
        )

    def revert_episode(self, episode_object: "TVEpisode"):
        """Restore the episodes of a failed download to their original state"""
        if not settings.USE_FAILED_DOWNLOADS:
            return

        sql_results = self.failed_db.select(
            "SELECT episode, old_status FROM history WHERE showid = ? AND season = ?", [episode_object.show.indexerid, episode_object.season]
        )

        history_eps = {res["episode"]: res for res in sql_results}

        try:
            logger.info(f"Reverting episode ({episode_object.season}, {episode_object.episode}): {episode_object.episode}")
            with episode_object.lock:
                if episode_object.episode in history_eps:
                    logger.info("Found in history")
                    episode_object.status = history_eps[episode_object.episode]["old_status"]
                else:
                    logger.debug("Episode don't have a previous snatched status to revert. Setting it back to WANTED")
                    episode_object.status = WANTED
                    episode_object.save_to_db()

        except EpisodeNotFoundException as error:
            logger.warning(f"Unable to create episode, please set its status manually: {error}")

    def mark_failed(self, episode_object: "TVEpisode"):
        """
        Mark an episode_object as failed

        :param episode_object: Episode object to mark as failed
        :return: empty string
        """
        if not settings.USE_FAILED_DOWNLOADS:
            return

        logger.info(f"Marking episode as bad: [{episode_object.pretty_name}]")
        try:
            with episode_object.lock:
                quality = Quality.splitCompositeStatus(episode_object.status)[1]
                episode_object.status = Quality.compositeStatus(FAILED, quality)
                episode_object.save_to_db()

        except EpisodeNotFoundException as error:
            logger.warning(f"Unable to get episode, please set its status manually: {error}")

        (release, provider) = self.find_release(episode_object)
        if release:
            self.log_failed(episode_object, release, provider)

        self.revert_episode(episode_object)

    def remove_snatch(self, release: str, size: int, provider: str):
        """
        Remove a snatch from history

        :param release: release to delete
        :param size: Size of release
        :param provider: Provider to delete it from
        """
        self.failed_db.action('DELETE FROM history WHERE "release" = ? AND size = ? AND provider = ?', [self.prepare_failed_name(release), size, provider])

    def find_release(self, episode_object: "TVEpisode"):
        """
        Find releases in history by show ID and season.
        Return None for release if multiple found or no release found.
        """

        # Clear old snatches for this release if any exist
        # self.failed_db.action(
        #     "DELETE FROM history WHERE showid = {0} AND season = {1} AND episode = {2}"
        #     " AND date < (SELECT max(date) FROM history WHERE showid = {0} AND season = {1} AND episode = {2})".format
        #     (episode_object.show.indexerid, episode_object.season, episode_object.episode)
        # )

        # Search for release in snatch history
        results = self.failed_db.select(
            'SELECT "release", provider, date FROM history WHERE showid = ? AND season = ? AND episode = ?',
            [episode_object.show.indexerid, episode_object.season, episode_object.episode],
        )

        for result in results:
            release = str(result["release"])
            provider = str(result["provider"])
            date = result["date"]

            # Clear any incomplete snatch records for this release if any exist
            self.failed_db.action('DELETE FROM history WHERE "release" = ? AND date <> ?', [release, date])

            # Found a previously failed release
            logger.debug(f"Failed release found for season ({episode_object.season}): ({result['release']})")
            return release, provider

        # Release was not found
        logger.debug(f"No releases found for season ({episode_object.season}) of ({episode_object.show.indexerid})")
        return None, None



from sickchill import settings
from sickchill.oldbeard.event_queue import Events


class Restart(object):
    def __init__(self):
        pass

    @staticmethod
    def restart(pid):
        if str(pid) != str(settings.PID):
            return False

        settings.events.put(Events.SystemEvent.RESTART)
        settings.restarting = True

        return True



import markdown2

from .. import logger
from ..oldbeard import helpers
from .common import PageTemplate
from .home import Home
from .routes import Route


@Route("/changes(/?.*)", name="changelog")
class HomeChangeLog(Home):
    def index(self, *args, **kwargs):
        # noinspection PyBroadException
        try:
            changes = helpers.getURL("https://raw.githubusercontent.com/SickChill/sickchill/master/CHANGES.md", session=helpers.make_session(), returns="text")
        except Exception:
            logger.debug("Could not load changes from repo, giving a link!")
            changes_url = "https://raw.githubusercontent.com/SickChill/sickchill/master/CHANGES.md"
            changes = _("Could not load changes from the repo. [Click here for CHANGES.md]") + f"({changes_url})"

        t = PageTemplate(rh=self, filename="markdown.mako")
        data = markdown2.markdown(changes if changes else _("The was a problem connecting to github, please refresh and try again"), extras=["header-ids"])

        return t.render(
            title=_("Changelog"),
            header=_("Changelog"),
            topmenu="system",
            data=data,
            controller="changes",
            action="index",
        )



import logging
import os
import string
import sys
import unittest
from pathlib import Path

from sickchill.oldbeard import browser


class BrowserTestAll(unittest.TestCase):
    """
    Test methods in `oldbeard.browser`
    """

    def setUp(self):
        self.here = Path(__file__)

    @unittest.skipUnless(os.name == "nt", "Test on Windows only")
    def test_get_win_drives(self):
        """
        Test getWinDrives
        """
        drives = browser.get_windows_drives()
        self.assertIsNotNone(drives)
        self.assertIn(f"{Path.cwd().drive}{os.sep}", drives, f"Drives[{drives}]")

    def test_folders_at_path(self):
        """
        Test foldersAtPath
        """
        test_list = browser.folders_at_path(self.here / "not_a_real_path")
        assert test_list[0]["currentPath"] == str(self.here.parent)

        test_list = browser.folders_at_path(Path("//"))
        if os.name == "nt":
            assert test_list[0]["currentPath"] == "My Computer", test_list[0]
            drives = browser.get_windows_drives()
            assert len(drives)
            assert len(test_list[1:])
            for item in test_list[1:]:
                if item["path"].upper()[0] in string.ascii_uppercase:
                    assert item["path"][:3] in drives, (item["path"], drives)
        else:
            assert test_list[0]["currentPath"] == "/"

        test_list = browser.folders_at_path(self.here, include_parent=True)
        assert test_list[0]["currentPath"] == str(self.here.parent)
        assert test_list[1]["name"] == ".."
        assert test_list[1]["path"] == str(self.here.parent.parent)


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger(__name__).setLevel(logging.DEBUG)

    SUITE = unittest.TestLoader().loadTestsFromTestCase(BrowserTestAll)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



import datetime
import re
from xml.etree import ElementTree

from babelfish import Country

import sickchill
from sickchill.helper.common import dateFormat
from sickchill.oldbeard import helpers

from ... import logger
from . import generic


class KODIMetadata(generic.GenericMetadata):
    """
    Metadata generation class for KODI 12+.

    The following file structure is used:

    show_root/tvshow.nfo                    (show metadata)
    show_root/fanart.jpg                    (fanart)
    show_root/poster.jpg                    (poster)
    show_root/banner.jpg                    (banner)
    show_root/Season ##/filename.ext        (*)
    show_root/Season ##/filename.nfo        (episode metadata)
    show_root/Season ##/filename-thumb.jpg  (episode thumb)
    show_root/season##-poster.jpg           (season posters)
    show_root/season##-banner.jpg           (season banners)
    show_root/season-all-poster.jpg         (season all poster)
    show_root/season-all-banner.jpg         (season all banner)
    """

    def __init__(
        self,
        show_metadata=False,
        episode_metadata=False,
        fanart=False,
        poster=False,
        banner=False,
        episode_thumbnails=False,
        season_posters=False,
        season_banners=False,
        season_all_poster=False,
        season_all_banner=False,
    ):
        super().__init__(
            show_metadata, episode_metadata, fanart, poster, banner, episode_thumbnails, season_posters, season_banners, season_all_poster, season_all_banner
        )

        self.name = "KODI"

        self.poster_name = "poster.jpg"
        self.season_all_poster_name = "season-all-poster.jpg"

        # web-ui metadata template
        self.eg_show_metadata = "tvshow.nfo"
        self.eg_episode_metadata = "Season##\\<i>filename</i>.nfo"
        self.eg_fanart = "fanart.jpg"
        self.eg_poster = "poster.jpg"
        self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>-thumb.jpg"
        self.eg_season_posters = "season##-poster.jpg"
        self.eg_season_banners = "season##-banner.jpg"
        self.eg_season_all_poster = "season-all-poster.jpg"
        self.eg_season_all_banner = "season-all-banner.jpg"

    @staticmethod
    def _split_info(info_string):
        return {x.strip().title() for x in re.sub(r"[,/]+", "|", info_string).split("|") if x.strip()}

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for an KODI-style tvshow.nfo and
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        indexer_show = sickchill.indexer.series(show_obj)
        if not indexer_show:
            logger.info("Unable to find show with id {} on {}, skipping it".format(show_obj.indexerid, show_obj.idxr.name))
            return False

        # check for title and id
        if not (getattr(indexer_show, "seriesName", None) and getattr(indexer_show, "id", None)):
            logger.info("Incomplete info for show with id {} on {}, skipping it".format(show_obj.indexerid, show_obj.idxr.name))
            return False

        tvshow_element = ElementTree.Element("tvshow")

        title_element = ElementTree.SubElement(tvshow_element, "title")
        title_element.text = indexer_show.seriesName

        if getattr(indexer_show, "rating", None):
            mpaa_element = ElementTree.SubElement(tvshow_element, "mpaa")
            mpaa_element.text = str(indexer_show.rating)

        if getattr(indexer_show, "siteRating", None):
            rating_element = ElementTree.SubElement(tvshow_element, "rating")
            rating_element.text = str(indexer_show.siteRating)

        if getattr(indexer_show, "firstAired", None):
            try:
                year_text = str(datetime.datetime.strptime(indexer_show.firstAired, dateFormat).year)
                if year_text:
                    year = ElementTree.SubElement(tvshow_element, "year")
                    year.text = year_text
            except Exception:
                pass

        if getattr(indexer_show, "overview", None):
            plot_element = ElementTree.SubElement(tvshow_element, "plot")
            plot_element.text = indexer_show.overview

        if getattr(indexer_show, "id", None):
            episodeguide_element = ElementTree.SubElement(tvshow_element, "episodeguide")
            episodeguide_url_element = ElementTree.SubElement(episodeguide_element, "url")
            episodeguide_url_element.attrib = {"post": "yes", "cache": "auth.json"}
            episodeguide_url_element.text = show_obj.idxr.episode_guide_url(show_obj)

            indexerid_element = ElementTree.SubElement(tvshow_element, "id")
            indexerid_element.text = str(indexer_show.id)

            indexerid_element = ElementTree.SubElement(tvshow_element, "tvdbid")
            indexerid_element.text = str(indexer_show.id)

        if getattr(indexer_show, "genre", None) and isinstance(indexer_show.genre, list):
            for genre in indexer_show.genre:
                genre_element = ElementTree.SubElement(tvshow_element, "genre")
                genre_element.text = genre

        if show_obj.imdb_info.get("country_codes"):
            for country in self._split_info(show_obj.imdb_info["country_codes"]):
                try:
                    country_name = Country(country.upper()).name.title()
                except Exception:
                    continue

                country_element = ElementTree.SubElement(tvshow_element, "country")
                country_element.text = country_name

        if getattr(indexer_show, "firstAired", None):
            premiered_element = ElementTree.SubElement(tvshow_element, "premiered")
            premiered_element.text = indexer_show.firstAired

        if getattr(indexer_show, "network", None):
            studio_element = ElementTree.SubElement(tvshow_element, "studio")
            studio_element.text = indexer_show.network.strip()

        data = show_obj.idxr.actors(indexer_show)
        if data:
            for actor in data:
                actor_element = ElementTree.SubElement(tvshow_element, "actor")

                if "name" in actor and actor["name"].strip():
                    actor_name_element = ElementTree.SubElement(actor_element, "name")
                    actor_name_element.text = actor["name"].strip()
                else:
                    continue

                if "role" in actor and actor["role"].strip():
                    actor_role_element = ElementTree.SubElement(actor_element, "role")
                    actor_role_element.text = actor["role"].strip()

                if "image" in actor and actor["image"].strip():
                    actor_thumb_element = ElementTree.SubElement(actor_element, "thumb")
                    actor_thumb_element.text = show_obj.idxr.complete_image_url(actor["image"])

        # Make it purdy
        helpers.indentXML(tvshow_element)

        return ElementTree.ElementTree(tvshow_element)

    def _ep_data(self, episode_object):
        """
        Creates an elementTree XML structure for an KODI-style episode.nfo and
        returns the resulting data object.
            show_obj: a TVEpisode instance to create the NFO for
        """

        episodes_to_write = [episode_object] + episode_object.related_episodes

        indexer_show = episode_object.idxr.series_from_episode(episode_object)
        if not indexer_show:
            logger.info("Unable to find {} on {} while creating meta files, skipping".format(episode_object.show.indexerid, episode_object.idxr.name))
            return

        if len(episodes_to_write) > 1:
            root_node = ElementTree.Element("kodimultiepisode")
        else:
            root_node = ElementTree.Element("episodedetails")

        # write an NFO containing info for all matching episodes
        for current_episode in episodes_to_write:
            indexer_episode = current_episode.idxr.episode(current_episode)
            if not indexer_episode:
                logger.info(
                    "Metadata writer is unable to find episode {0:d}x{1:d} of {2} on {3}..."
                    "has it been removed? Should I delete from db?".format(
                        current_episode.season, current_episode.episode, current_episode.show.name, episode_object.idxr.name
                    )
                )

                return None

            if episode_object.airdate != datetime.date.min and not indexer_episode.get("firstAired"):
                indexer_episode["firstAired"] = str(episode_object.airdate)

            if not indexer_episode.get("episodeName"):
                logger.debug("Not generating nfo because the ep has no title")
                return None

            if not indexer_episode.get("firstAired"):
                logger.debug("Not generating nfo because the ep has no airdate")
                return None

            logger.debug(f"Creating metadata for episode {episode_object.episode_number}")

            if len(episodes_to_write) > 1:
                episode_element = ElementTree.SubElement(root_node, "episodedetails")
            else:
                episode_element = root_node

            if indexer_episode.get("episodeName"):
                title_element = ElementTree.SubElement(episode_element, "title")
                title_element.text = indexer_episode["episodeName"]

            if getattr(indexer_show, "seriesName", None):
                showtitle_element = ElementTree.SubElement(episode_element, "showtitle")
                showtitle_element.text = indexer_show.seriesName

            season_element = ElementTree.SubElement(episode_element, "season")
            season_element.text = str(current_episode.season)

            episode_number_element = ElementTree.SubElement(episode_element, "episode")
            episode_number_element.text = str(current_episode.episode)

            if indexer_episode.get("id"):
                uniqueid_element = ElementTree.SubElement(episode_element, "uniqueid")
                uniqueid_element.text = str(indexer_episode["id"])

            if current_episode.airdate != datetime.date.min:
                aired_element = ElementTree.SubElement(episode_element, "aired")
                aired_element.text = str(current_episode.airdate)

            if indexer_episode.get("overview"):
                plot_element = ElementTree.SubElement(episode_element, "plot")
                plot_element.text = indexer_episode["overview"]

            if current_episode.season and getattr(indexer_show, "runtime", None):
                runtime_element = ElementTree.SubElement(episode_element, "runtime")
                runtime_element.text = indexer_show.runtime

            if indexer_episode.get("airedSeason"):
                displayseason_element = ElementTree.SubElement(episode_element, "displayseason")
                displayseason_element.text = str(indexer_episode["airedSeason"])

            if indexer_episode.get("airedEpisodeNumber"):
                displayepisode_element = ElementTree.SubElement(episode_element, "displayepisode")
                displayepisode_element.text = str(indexer_episode["airedEpisodeNumber"])

            if indexer_episode.get("filename"):
                thumb_element = ElementTree.SubElement(episode_element, "thumb")
                thumb_element.text = episode_object.idxr.complete_image_url(indexer_episode["filename"])

            if indexer_episode.get("rating") is not None:
                rating_element = ElementTree.SubElement(episode_element, "rating")
                rating_element.text = str(indexer_episode["rating"])

            if indexer_episode.get("writers") and isinstance(indexer_episode["writers"], list):
                for writer in indexer_episode["writers"]:
                    writer_element = ElementTree.SubElement(episode_element, "credits")
                    writer_element.text = writer

            if indexer_episode.get("directors") and isinstance(indexer_episode["directors"], list):
                for director in indexer_episode["directors"]:
                    director_element = ElementTree.SubElement(episode_element, "director")
                    director_element.text = director

            if indexer_episode.get("guestStars") and isinstance(indexer_episode["guestStars"], list):
                for actor in indexer_episode["guestStars"]:
                    actor_element = ElementTree.SubElement(episode_element, "actor")
                    actor_name_element = ElementTree.SubElement(actor_element, "name")
                    actor_name_element.text = actor

            for actor in episode_object.idxr.actors(indexer_show):
                actor_element = ElementTree.SubElement(episode_element, "actor")

                if "name" in actor and actor["name"].strip():
                    actor_name_element = ElementTree.SubElement(actor_element, "name")
                    actor_name_element.text = actor["name"].strip()
                else:
                    continue

                if "role" in actor and actor["role"].strip():
                    actor_role_element = ElementTree.SubElement(actor_element, "role")
                    actor_role_element.text = actor["role"].strip()

                if "image" in actor and actor["image"].strip():
                    actor_thumb_element = ElementTree.SubElement(actor_element, "thumb")
                    actor_thumb_element.text = episode_object.idxr.complete_image_url(actor["image"])

        # Make it purdy
        helpers.indentXML(root_node)

        return ElementTree.ElementTree(root_node)


# present a standard "interface" from the module
metadata_class = KODIMetadata



import re
from typing import Dict, List, TYPE_CHECKING
from urllib.parse import urljoin

from sickchill import logger, settings
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.oldbeard.show_name_helpers import allPossibleShowNames

from .TorrentProvider import TorrentProvider

if TYPE_CHECKING:
    from sickchill.tv import TVEpisode


class FrenchTorrentProvider(TorrentProvider):
    def __init__(self, name, url):
        super().__init__(name)

        self.proper_strings = ["PROPER", "REPACK"]
        self.cache = tvcache.TVCache(self)

        self.public = True
        self.minseed = 0
        self.minleech = 0

        self._original_url = url
        self._custom_url = None
        self._used_url = None
        self._recheck_url = True

    def _retrieve_dllink_from_url(self, inner_url):
        data = self.get_url(urljoin(self.url, inner_url), returns="text")
        regex = r".*?function\s+redirect\(\).+?= '([^']+)'"

        magnet_classes = ["btn-magnet", "magnet-btn"]
        file_classes = ["btn-download", "download-btn"]

        # Always prefer magnets, unless using black hole.
        check_classes = magnet_classes + file_classes

        if settings.TORRENT_METHOD == "blackhole":
            check_classes.reverse()

        with BS4Parser(data) as html:
            scripts = html.head.find_all("script")
            if len(scripts):
                script = scripts[-1].get("src")
                matches = re.match(regex, script, re.S)
                if matches:
                    return urljoin(self.url, matches[1])

            download_btns = html.find_all("div", {"class": check_classes})
            for btn in download_btns:
                link = btn.find("a").get("href")
                if link.startswith("javascript"):
                    return ""
                if link.startswith("magnet"):
                    return link
                else:
                    return urljoin(self.url, link)

        return ""

    @property
    def custom_url(self):
        return self._custom_url

    @custom_url.setter
    def custom_url(self, url):
        if self._custom_url != url:
            self._custom_url = url
            self._recheck_url = True

    @property
    def url(self):
        if self._recheck_url:
            if self.custom_url:
                if self.valid_url(self.custom_url):
                    self._used_url = self.custom_url
                else:
                    logger.warning("Invalid custom url set, please check your settings")

            self._used_url = self._original_url

        return self._used_url

    @url.setter
    def url(self, url):
        self._used_url = url

    def get_season_search_strings(self, episode: "TVEpisode") -> List[Dict]:
        search_string = {"Season": set()}
        for show_name in allPossibleShowNames(episode.show, season=episode.scene_season):
            season = int(episode.scene_season)
            if episode.show.air_by_date or episode.show.sports:
                year = str(episode.airdate).split("-")[0]
                season_string = f"{show_name} {year}"
            elif episode.show.anime:
                season_string = f"{show_name} Saison"  # ignore season number to get all seasons in all formats
            else:
                season_string = f"{show_name} Saison {season:0d}"

            search_string["Season"].add(season_string)

        return [search_string]



"""
This is the global flask application under which all web interfaces for sickchill media types run on
"""

import threading
from logging.config import dictConfig

from flask import Flask

from .config import blueprint as config_blueprint
from .movies import blueprint as movies_blueprint
from .shows import blueprint as shows_blueprint


class FlaskServer(threading.Thread):
    """Flask application class to set up the sickchill flask webserver interface"""

    def __init__(self, host, port):
        super().__init__(name="FLASK")
        self.app = None
        self.host = host
        self.port = port
        self.daemon = True
        self.alive = True

    def run(self):
        """Run the flask server"""
        self.app = Flask("SickFlask", template_folder="frontend/templates", static_folder="frontend/static", static_url_path="/static")
        self.app.config.from_object(DevelopmentConfig)

        self.app.register_blueprint(config_blueprint)
        self.app.register_blueprint(shows_blueprint)
        self.app.register_blueprint(movies_blueprint)
        routes = "\n\t\t".join(f"{rule}" for rule in self.app.url_map.iter_rules())
        self.app.logger.debug(f"{self.app.name} Route:\n\t\t{routes}")

        self.app.run(host=self.host, port=self.port, use_reloader=False)

    def stop(self):
        """Stop the flask server"""
        self.alive = False
        del self.app


class BaseConfig:
    """
    Base config class
    """

    DEBUG = True
    TESTING = False


class ProductionConfig(BaseConfig):
    """
    Production specific config
    """

    DEBUG = False


class DevelopmentConfig(BaseConfig):
    """
    Development environment specific configuration
    """

    DEBUG = True
    TESTING = True


if __name__ == "__main__":
    dictConfig(
        {
            "version": 1,
            "formatters": {
                "default": {
                    "format": "[%(asctime)s] %(levelname)s in %(module)s: %(message)s",
                }
            },
            "handlers": {"wsgi": {"class": "logging.StreamHandler", "stream": "ext://flask.logging.wsgi_errors_stream", "formatter": "default"}},
            "root": {"level": "DEBUG", "handlers": ["wsgi"]},
        }
    )



import argparse
import gettext
import logging
import os
import re
import sys
from importlib.metadata import Distribution, PackageNotFoundError
from pathlib import Path
from typing import Union

logger = logging.getLogger(__file__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())

# locale_dir = sickchill_dir / "locale"
pid_file: Union[Path, None] = None


sickchill_dir = Path(__file__).parent
locale_dir = sickchill_dir / "locale"
pyproject_file = sickchill_dir.parent / "pyproject.toml"
git_folder = sickchill_dir.parent / ".git"


def setup_gettext(language: str = None) -> None:
    languages = [language] if language else None
    if not [key for key in ("LANGUAGE", "LC_ALL", "LC_MESSAGES", "LANG") if os.environ.get(key)]:
        os.environ["LC_MESSAGES"] = "en_US.UTF-8"
    gt = gettext.translation("messages", locale_dir, languages=languages, fallback=True)
    gt.install(names=["ngettext"])


def maybe_daemonize():
    """
    Fork off as a daemon
    """
    if sys.platform == "win32":
        return

    args = argparse.ArgumentParser(add_help=False)
    args.add_argument("-d", "--daemon", action="store_true")
    args.add_argument("--pidfile")
    args, extra = args.parse_known_args(sys.argv)

    if not args.daemon:
        return

    global pid_file
    if args.pidfile:
        pid_file = Path(args.pidfile).resolve()

    if pid_file:
        if pid_file.is_file():
            # If the pid file already exists, SickChill may still be running, so exit
            raise SystemExit(f"PID file: {pid_file} already exists. Exiting.")
        pid_dir = pid_file.parent
        if not os.access(pid_dir, os.F_OK):
            raise SystemExit(f"PID dir: {pid_dir} doesn't exist. Exiting.")
        if not os.access(pid_dir, os.W_OK):
            raise SystemExit(f"PID dir: {pid_dir} must be writable (write permissions). Exiting.")

    # Make a non-session-leader child process
    try:
        pid = os.fork()  # @UndefinedVariable - only available in UNIX
        if pid != 0:
            os._exit(0)  # noqa
    except OSError as error:
        raise SystemExit(f"fork #1 failed: {error}\n")

    os.setsid()  # @UndefinedVariable - only available in UNIX

    # https://github.com/SickChill/SickChill/issues/2969
    # http://www.microhowto.info/howto/cause_a_process_to_become_a_daemon_in_c.html#idp23920
    # https://www.safaribooksonline.com/library/view/python-cookbook/0596001673/ch06s08.html
    # Previous code simply set the umask to whatever it was because it was ANDing instead of OR-ing
    # Daemons traditionally run with umask 0 anyways and this should not have repercussions
    os.umask(0)

    # Make the child a session-leader by detaching from the terminal
    try:
        pid = os.fork()
        if pid != 0:
            os._exit(0)  # noqa
    except OSError as error:
        raise SystemExit(f"fork #2 failed: {error}\n")

    # Write pid
    if pid_file:
        pid = os.getpid()

        logger.info(f"Writing PID: {pid} to {pid_file}\n")
        try:
            pid_file.write_text(f"{pid}\n")
        except EnvironmentError as error:
            raise SystemExit(f"Unable to write PID file: {pid_file} Error {error}")

    # Redirect all output
    sys.stdout.flush()
    sys.stderr.flush()

    devnull = getattr(os, "devnull", "/dev/null")
    stdin = open(devnull)
    stdout = open(devnull, "a+")
    stderr = open(devnull, "a+")

    os.dup2(stdin.fileno(), getattr(sys.stdin, "device", sys.stdin).fileno())
    os.dup2(stdout.fileno(), getattr(sys.stdout, "device", sys.stdout).fileno())
    os.dup2(stderr.fileno(), getattr(sys.stderr, "device", sys.stderr).fileno())


def remove_pid_file():
    """
    Remove pid file

    :return:
    """
    try:
        if pid_file and pid_file.exists():
            pid_file.unlink()
    except EnvironmentError:
        pass


def get_distribution() -> Union[Distribution, None]:
    try:
        distribution = Distribution.from_name(__package__)
    except PackageNotFoundError:
        return None

    return distribution


def check_installed() -> bool:
    if pyproject_file.is_file() or git_folder.is_dir():
        return False

    return get_distribution() is not None


def get_current_version() -> str:
    fallback_version = "0.0.0"
    matcher = re.compile(r'\s*version\s*=\s*["\']([.0-9a-z-+]+)["\']\s*$')
    if pyproject_file.is_file():
        for line in pyproject_file.open():
            match = matcher.match(line)
            if match:
                return match.group(1)
        return fallback_version

    try:
        return get_distribution().version
    except PackageNotFoundError:
        return fallback_version



import traceback

from sickchill import logger
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider, tvcache.RSSTorrentMixin):
    def __init__(self):
        super().__init__("Nyaa")

        self.public = True
        self.supports_absolute_numbering = True
        self.anime_only = True

        self.url = "https://nyaa.si"
        self.custom_url = None

        self.minseed = 0
        self.minleech = 0
        self.confirmed = False

        self.size_units = ["BYTES", "KIB", "MIB", "GIB", "TIB", "PIB"]
        self.cache = tvcache.TVCache(self, min_time=20)  # only poll Nyaa every 20 minutes max

    def search(self, search_strings):
        results = []
        if self.show and not self.show.is_anime:
            return results

        if self.custom_url:
            if self.invalid_url(self.custom_url):
                logger.warning(_("Invalid custom url: {0}").format(self.custom_url))
                return results

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}").format(mode=mode))
            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}").format(search_string=search_string))

                search_params = {
                    "page": "rss",
                    "c": "1_0",  # Category: All anime
                    "s": "id",  # Sort by: 'id'=Date / 'size' / 'name' / 'seeders' / 'leechers' / 'downloads'
                    "o": "desc",  # Sort direction: asc / desc
                    "f": ("0", "2")[self.confirmed],  # Quality filter: 0 = None / 1 = No Remakes / 2 = Trusted Only
                }
                if mode != "RSS":
                    search_params["q"] = search_string

                results = []
                data = self.get_url(self.custom_url or self.url, params=search_params)
                if not data:
                    logger.debug("No data was returned from the provider")
                    continue

                with BS4Parser(data, language="xml") as html:
                    for item in html.find_all("item"):
                        try:
                            result = self.parse_feed_item(item, self.url, size_units=self.size_units)
                            if result:
                                if result["seeders"] < self.minseed or result["leechers"] < self.minleech:
                                    if mode != "RSS":
                                        logger.debug(
                                            f"Discarding torrent because it doesn't meet the minimum seeders or leechers: {result['title']} (S:{result['seeders']} L:{result['leechers']})"
                                        )
                                    continue
                                items.append(result)
                        except Exception as error:
                            logger.debug(f"Error parsing results: {error}")
                            logger.debug(traceback.format_exc())
                            continue

            # For each search mode sort all the items by seeders
            items.sort(key=lambda d: d.get("seeders", 0), reverse=True)
            results += items

        return results



from flask import Blueprint, render_template

from sickchill import logger, settings

blueprint = Blueprint("config", __name__, template_folder="templates", static_folder="static", url_prefix="/config")


@blueprint.route("/")
def config():
    logger.info("Loading config page")
    return render_template("config.html")



import os
from urllib.parse import urlencode
from urllib.request import Request, urlopen

import requests

from sickchill import logger, settings, tv


class Notifier(object):
    def notify_snatch(self, ep_name):
        pass

    def notify_download(self, ep_name):
        pass

    def notify_subtitle_download(self, ep_name, lang):
        pass

    def notify_update(self, new_version):
        pass

    def notify_login(self, ipaddress=""):
        pass

    @staticmethod
    def update_library(episode_object):
        # Values from config

        if not (settings.USE_PYTIVO and settings.PYTIVO_HOST):
            return False

        # There are two more values required, the container and file.
        #
        # container: The share name, show name and season
        #
        # file: The file name
        #
        # Some slicing and dicing of variables is required to get at these values.
        #
        # There might be better ways to arrive at the values, but this is the best I have been able to
        # come up with.
        #

        # Calculated values
        if isinstance(episode_object, tv.TVEpisode):
            show_path = episode_object.show.location
            show_name = episode_object.show.name
            show_and_season_path = os.path.dirname(episode_object.location)
            absolute_file_path = episode_object.location
        else:
            # This is a TVShow
            show_path = episode_object.location
            show_name = episode_object.name
            show_and_season_path = os.path.dirname(episode_object.location)
            absolute_file_path = episode_object.location

        # Some show names have colons in them which are illegal in a path location, so strip them out.
        # (Are there other characters?)
        show_name = show_name.replace(":", "")

        root = show_path.replace(show_name, "")
        show_and_season_path = show_and_season_path.replace(root, "")

        container = settings.PYTIVO_SHARE_NAME + "/" + show_and_season_path
        filename = "/" + absolute_file_path.replace(root, "")

        # Finally create the url and make request
        requestUrl = (
            "http://"
            + settings.PYTIVO_HOST
            + "/TiVoConnect?"
            + urlencode({"Command": "Push", "Container": container, "File": filename, "tsn": settings.PYTIVO_TIVO_NAME})
        )

        logger.debug("pyTivo notification: Requesting " + requestUrl)

        request = Request(requestUrl)

        try:
            urlopen(request)
        except requests.exceptions.RequestException as error:
            if hasattr(error, "reason"):
                logger.exception(f"pyTivo notification: Error, failed to reach a server - {error.reason}")
                return False
            elif hasattr(error, "code"):
                logger.exception(f"pyTivo notification: Error, the server couldn't fulfill the request - {error.code}")
            return False
        except Exception as error:
            logger.exception(f"PYTIVO: Unknown exception: {error}")
            return False
        else:
            logger.info("pyTivo notification: Successfully requested transfer of file")
            return True



import re
import string
from pathlib import Path

from . import aniDBfileInfo as fileInfo
from .aniDBerrors import AniDBIncorrectParameterError
from .aniDBfileInfo import read_anidb_xml
from .aniDBmapper import AniDBMapper
from .aniDBtvDBmaper import TvDBMap


class AniDBabstractObject(object):
    def __init__(self, aniDB, load=False):
        self.laoded = False
        self.aniDB = None
        self.rawData = None
        self.log = self._fake_log

        self.set_connection(aniDB)
        if load:
            self.load_data()

    def set_connection(self, anidb):
        self.aniDB = anidb
        if self.aniDB:
            self.log = self.aniDB.log
        else:
            self.log = self._fake_log

    def _fake_log(self, x=None):
        pass

    def _fill(self, dataline):
        for key in dataline:
            try:
                tmp_list = dataline[key].split("'")
                if len(tmp_list) > 1:
                    new_list = []
                    for i in tmp_list:
                        try:
                            new_list.append(int(i))
                        except Exception:
                            new_list.append(i)
                    self.__dict__[key] = new_list
                    continue
            except Exception:
                pass
            try:
                self.__dict__[key] = int(dataline[key])
            except Exception:
                self.__dict__[key] = dataline[key]
            key = property(lambda x: dataline[key])

    def __getattr__(self, name):
        try:
            return object.__getattribute__(self, name)
        except Exception:
            return None

    def _build_names(self):
        names = []
        names = self._easy_extend(names, self.english_name)
        names = self._easy_extend(names, self.short_name_list)
        names = self._easy_extend(names, self.synonym_list)
        names = self._easy_extend(names, self.other_name)

        self.allNames = names

    @staticmethod
    def _easy_extend(initialList, item):
        if item:
            if isinstance(item, list):
                initialList.extend(item)
            elif isinstance(item, str):
                initialList.append(item)

        return initialList

    def load_data(self):
        return False

    def add_notification(self):
        """
        type - Type of notification: type=>  0=all, 1=new, 2=group, 3=complete
        priority - low = 0, medium = 1, high = 2 (unconfirmed)

        """
        if self.aid:
            self.aniDB.notifyadd(aid=self.aid, type=1, priority=1)

    def del_notification(self):
        """
        type - Type of notification: type=>  0=all, 1=new, 2=group, 3=complete
        priority - low = 0, medium = 1, high = 2 (unconfirmed)

        """
        if self.aid:
            self.aniDB.notifydel(aid=self.aid, type=1, priority=1)


class Anime(AniDBabstractObject):
    def __init__(self, aniDB, cache_dir: Path, name=None, aid=None, tvdbid=None, paramsA=None, autoCorrectName=False, load=False):
        if not cache_dir.is_dir():
            raise
        self.cache_dir = cache_dir / "anime"
        self.cache_dir = self.cache_dir.absolute()

        if not self.cache_dir.is_dir():
            self.cache_dir.mkdir()

        self.mapper = AniDBMapper()

        self.tvDBMap = TvDBMap(cache_dir=self.cache_dir)
        self.allAnimeXML = None

        self.name = name
        self.aid = aid
        self.tvdb_id = tvdbid
        self.release_groups = []

        if self.tvdb_id and not self.aid:
            self.aid = self.tvDBMap.get_anidb_for_tvdb(self.tvdb_id)

        if not (self.name or self.aid):
            raise AniDBIncorrectParameterError("No aid or name available")

        if not self.aid:
            self.aid = self._get_aid_from_xml(self.name)
        if not self.name or autoCorrectName:
            self.name = self._get_name_from_xml(self.aid)

        if not (self.name or self.aid):
            raise ValueError

        if not self.tvdb_id:
            self.tvdb_id = self.tvDBMap.get_tvdb_for_anidb(self.aid)

        if not paramsA:
            self.bitCode = "b2f0e0fc000000"
            self.params = self.mapper.getAnimeCodesA(self.bitCode)
        else:
            self.paramsA = paramsA
            self.bitCode = self.mapper.getAnimeBitsA(self.paramsA)

        super().__init__(aniDB, load)

    def load_data(self):
        """load the data from anidb"""

        if not (self.name or self.aid):
            raise ValueError

        self.rawData = self.aniDB.anime(aid=self.aid, aname=self.name, amask=self.bitCode)
        if self.rawData.datalines:
            self._fill(self.rawData.datalines[0])
            self._builPreSequal()
            self.laoded = True

    def get_groups(self):
        if not self.aid:
            return []
        self.rawData = self.aniDB.groupstatus(aid=self.aid)
        for line in self.rawData.datalines:
            self.release_groups.append({"name": line["name"], "rating": line["rating"], "range": line["episode_range"]})
        return sorted(self.release_groups, key=lambda x: x["name"].lower())

    # TODO: refactor and use the new functions in anidbFileinfo
    def _get_aid_from_xml(self, name):
        if not self.allAnimeXML:
            self.allAnimeXML = read_anidb_xml(self.cache_dir)

        regex = re.compile(rf"( \(\d{4}\))|[{re.escape(string.punctuation)}]")  # remove any punctuation and e.g. ' (2011)'
        # regex = re.compile('[%s]'  % re.escape(string.punctuation)) # remove any punctuation and e.g. ' (2011)'
        name = regex.sub("", name.lower())
        lastAid = 0
        for element in self.allAnimeXML.iter():
            if element.get("aid", False):
                lastAid = int(element.get("aid"))
            if element.text:
                testname = regex.sub("", element.text.lower())

                if testname == name:
                    return lastAid
        return 0

    # TODO: refactor and use the new functions in anidbFileinfo
    def _get_name_from_xml(self, aid, onlyMain=True):
        if not self.allAnimeXML:
            self.allAnimeXML = read_anidb_xml(self.cache_dir)

        for anime in self.allAnimeXML.findall("anime"):
            if int(anime.get("aid", False)) == aid:
                for title in anime.iter():
                    currentLang = title.get("{http://www.w3.org/XML/1998/namespace}lang", False)
                    currentType = title.get("type", False)
                    if (currentLang == "en" and not onlyMain) or currentType == "main":
                        return title.text
        return ""

    def _builPreSequal(self):
        if self.related_aid_list and self.related_aid_type:
            try:
                for i in range(len(self.related_aid_list)):
                    if self.related_aid_type[i] == 2:
                        self.__dict__["prequal"] = self.related_aid_list[i]
                    elif self.related_aid_type[i] == 1:
                        self.__dict__["sequal"] = self.related_aid_list[i]
            except Exception:
                if self.related_aid_type == 2:
                    self.__dict__["prequal"] = self.related_aid_list
                elif self.str_related_aid_type == 1:
                    self.__dict__["sequal"] = self.related_aid_list


class Episode(AniDBabstractObject):
    def __init__(self, aniDB, number=None, epid=None, file_path: Path = None, fid=None, epno=None, paramsA=None, paramsF=None, load=False, calculate=False):
        self.mapper = AniDBMapper()
        self.epid = epid
        self.file_path = file_path
        self.fid = fid
        self.epno = epno

        if calculate:
            self.ed2k, self.size = self._calculate_file_stuff

        if not paramsA:
            self.bitCodeA = "C000F0C0"
            self.paramsA = self.mapper.getFileCodesA(self.bitCodeA)
        else:
            self.paramsA = paramsA
            self.bitCodeA = self.mapper.getFileBitsA(self.paramsA)

        if not paramsF:
            self.bitCodeF = "7FF8FEF8"
            self.paramsF = self.mapper.getFileCodesF(self.bitCodeF)
        else:
            self.paramsF = paramsF
            self.bitCodeF = self.mapper.getFileBitsF(self.paramsF)

        super().__init__(aniDB, load)

    def load_data(self):
        """load the data from anidb"""
        self.ed2k, self.size = self._calculate_file_stuff

        self.rawData = self.aniDB.file(
            fid=self.fid,
            size=self.size,
            ed2k=self.ed2k,
            aid=self.aid,
            aname=None,
            gid=None,
            gname=None,
            epno=self.epno,
            fmask=self.bitCodeF,
            amask=self.bitCodeA,
        )
        self._fill(self.rawData.datalines[0])
        self._build_names()
        self.laoded = True

    def add_to_mylist(self, status=None):
        """
        status:
        0    unknown    - state is unknown or the user doesn't want to provide this information (default)
        1    on hdd    - the file is stored on hdd
        2    on cd    - the file is stored on cd
        3    deleted    - the file has been deleted or is not available for other reasons (i.e. reencoded)

        """
        self.ed2k, self.size = self._calculate_file_stuff

        try:
            self.aniDB.mylistadd(size=self.size, ed2k=self.ed2k, state=status)
        except Exception as error:
            self.log(f"exception msg: {error}")
        else:
            # TODO: add the name or something
            self.log("Added the episode to anidb")

    @property
    def _calculate_file_stuff(self):
        if not (self.file_path and self.file_path.is_file()):
            return None, None

        if self.ed2k and self.size:
            return self.ed2k, self.size

        self.log("Calculating the ed2k. Please wait...")
        ed2k = fileInfo.get_file_hash(self.file_path)
        size = self.file_path.stat().st_size
        return ed2k, size



import datetime
import json
import os
import re
import traceback
from operator import itemgetter

import dateutil.parser
from tornado.web import HTTPError

import sickchill
from sickchill import logger, settings
from sickchill.helper import sanitize_filename, try_int
from sickchill.oldbeard import config, db, filters, helpers, ui
from sickchill.oldbeard.blackandwhitelist import short_group_names
from sickchill.oldbeard.common import Quality
from sickchill.oldbeard.trakt_api import TraktAPI
from sickchill.oldbeard.traktTrending import trakt_trending
from sickchill.show.recommendations.favorites import favorites
from sickchill.show.recommendations.imdb import imdb_popular
from sickchill.show.Show import Show
from sickchill.tv import TVShow
from sickchill.views.common import PageTemplate
from sickchill.views.home import Home
from sickchill.views.routes import Route


@Route("/addShows(/?.*)", name="addShows")
class AddShows(Home):
    def index(self):
        t = PageTemplate(rh=self, filename="addShows.mako")
        return t.render(
            title=_("Add Shows"),
            header=_("Add Shows"),
            topmenu="home",
            controller="addShows",
            action="index",
        )

    def sanitizeFileName(self):
        return sanitize_filename(self.get_body_argument("name"))

    def searchIndexersForShowName(self):
        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        self.set_header("Content-Type", "application/json")
        search_term = self.get_body_argument("search_term")
        search_terms = [search_term]  # get_arguments to make this a list of terms, we can probably add advanced searching here.
        lang = self.get_body_argument("lang", default=settings.INDEXER_DEFAULT_LANGUAGE)
        indexer = int(self.get_body_argument("indexer", default=settings.INDEXER_DEFAULT))
        exact = config.checkbox_to_value(self.get_body_argument("exact"))

        # If search term ends with what looks like a year, enclose it in ()
        matches = re.match(r"^(.+ |)([12][0-9]{3})$", search_term)
        if matches:
            search_terms.append("{0}({1})".format(matches.group(1), matches.group(2)))

        for term in search_terms:
            # If search term begins with an article, let's also search for it without
            matches = re.match(r"^(?:a|an|the) (.+)$", term, re.I)
            if matches:
                search_terms.append(matches.group(1))

        results = {}
        final_results = []

        # Query Indexers for each search term and build the list of results
        for index, indexer_object in sickchill.indexer:
            if indexer and indexer != index:
                continue

            logger.debug(
                _("Searching for Show with search term(s): {search_terms} on Indexer: {indexer_name} (exact: {exact})").format(
                    search_terms=search_terms, indexer_name=indexer_object.name, exact=exact
                )
            )
            for term in search_terms:
                # noinspection PyBroadException
                try:
                    indexer_results = indexer_object.search(term, language=lang, exact=exact)
                except Exception:
                    logger.debug(traceback.format_exc())
                    continue

                # add search results
                results.setdefault(index, []).extend(indexer_results)

        for index, shows in results.items():
            # noinspection PyUnresolvedReferences
            final_results.extend(
                {
                    (
                        sickchill.indexer[index].name,
                        index,
                        sickchill.indexer[index].show_url,
                        show["id"],
                        show["seriesName"],
                        show["firstAired"],
                        sickchill.tv.Show.find(settings.show_list, show["id"]) is not None,
                    )
                    for show in shows
                }
            )

        if exact in [True, "1"]:
            logger.debug(_("Filtering and sorting out excess results because exact match was checked"))
            final_results = [item for item in final_results if search_term.lower() in item[4].lower()]
            final_results.sort(key=itemgetter(4))
            final_results.sort(key=lambda x: x[4].lower().index(search_term.lower()))
            final_results.sort(key=lambda x: x[4].lower() == search_term.lower(), reverse=True)

        lang_id = sickchill.indexer[indexer].lang_dict[lang]
        return json.dumps({"results": final_results, "langid": lang_id, "success": len(final_results) > 0})

    def massAddTable(self):
        t = PageTemplate(rh=self, filename="home_massAddTable.mako")
        root_dirs = self.get_arguments("rootDir")
        if not root_dirs:
            return _("No folders selected.")

        dir_list = []

        main_db_con = db.DBConnection()
        for root_dir in root_dirs:
            # noinspection PyBroadException
            try:
                file_list = os.listdir(root_dir)
            except Exception:
                continue

            for cur_file in file_list:
                # noinspection PyBroadException
                try:
                    cur_path = os.path.normpath(os.path.join(root_dir, cur_file))
                    if not os.path.isdir(cur_path):
                        continue
                    # ignore Synology folders
                    # noinspection SpellCheckingInspection
                    if cur_file.lower() in ["#recycle", "@eadir"]:
                        continue
                except Exception:
                    continue

                cur_dir = {
                    "dir": cur_path,
                    "existing_info": (None, None, None),
                    "display_dir": "<b>" + os.path.dirname(cur_path) + os.sep + "</b>" + os.path.basename(cur_path),
                }

                dir_results = main_db_con.select("SELECT indexer_id FROM tv_shows WHERE location = ? LIMIT 1", [cur_path])

                cur_dir["added_already"] = bool(dir_results)

                dir_list.append(cur_dir)

                indexer_id = show_name = indexer = None
                for cur_provider in settings.metadata_provider_dict.values():
                    if not (indexer_id and show_name):
                        (indexer_id, show_name, indexer) = cur_provider.retrieveShowMetadata(cur_path)
                        if all((indexer_id, show_name, indexer)):
                            break

                if all((indexer_id, show_name, indexer)):
                    cur_dir["existing_info"] = (indexer_id, show_name, indexer)

                if indexer_id and Show.find(settings.show_list, indexer_id):
                    cur_dir["added_already"] = True
        return t.render(dirList=dir_list)

    def newShow(self, show_to_add=None, other_shows=None, search_string=None):
        """
        Display the new show page which collects a tvdb id, folder, and extra options and
        posts them to addNewShow
        """
        t = PageTemplate(rh=self, filename="addShows_newShow.mako")

        indexer, show_dir, indexer_id, show_name = self.split_extra_show(show_to_add)

        if indexer_id and indexer and show_name:
            use_provided_info = True
        else:
            use_provided_info = False

        if show_name:
            default_show_name = show_name
        elif show_dir:
            default_show_name = re.sub(r" \(\d{4}\)", "", os.path.basename(os.path.normpath(show_dir)).replace(".", " "))
        elif search_string:
            default_show_name = search_string
        else:
            default_show_name = ""

        # carry a list of other dirs if given
        if not other_shows:
            other_shows = []
        elif not isinstance(other_shows, list):
            other_shows = [other_shows]

        provided_indexer_id = int(indexer_id or 0)
        provided_indexer_name = show_name

        provided_indexer = int(indexer or settings.INDEXER_DEFAULT)

        return t.render(
            enable_anime_options=True,
            use_provided_info=use_provided_info,
            default_show_name=default_show_name,
            other_shows=other_shows,
            provided_show_dir=show_dir,
            provided_indexer_id=provided_indexer_id,
            provided_indexer_name=provided_indexer_name,
            provided_indexer=provided_indexer,
            whitelist=settings.WHITELIST_DEFAULT,
            blacklist=settings.BLACKLIST_DEFAULT,
            groups=[],
            title=_("New Show"),
            header=_("New Show"),
            topmenu="home",
            controller="addShows",
            action="newShow",
        )

    def trendingShows(self):
        """
        Display the new show page which collects a tvdb id, folder, and extra options and
        posts them to addNewShow
        """

        trakt_list = self.get_query_argument("traktList", default="anticipated").lower()

        # noinspection SpellCheckingInspection
        trakt_options = {
            "anticipated": _("Most Anticipated Shows"),
            "newshow": _("New Shows"),
            "newseason": _("Season Premieres"),
            "trending": _("Trending Shows"),
            "popular": _("Popular Shows"),
            "watched": _("Most Watched Shows"),
            "played": _("Most Played Shows"),
            "collected": _("Most Collected Shows"),
        }
        if settings.TRAKT_ACCESS_TOKEN:
            trakt_options["recommended"] = _("Recommended Shows")

        t = PageTemplate(rh=self, filename="addShows_trendingShows.mako")
        return t.render(
            title=trakt_options[trakt_list],
            header=trakt_options[trakt_list],
            traktList=trakt_list,
            trakt_options=trakt_options,
            controller="addShows",
            action="trendingShows",
        )

    def getTrendingShows(self):
        """
        Display the new show page which collects a tvdb id, folder, and extra options and posts them to addNewShow
        """
        t = PageTemplate(rh=self, filename="trendingShows.mako")

        trakt_list = self.get_query_argument("traktList", "").lower()

        # noinspection SpellCheckingInspection
        if trakt_list == "trending":
            page_url = "shows/trending"
        elif trakt_list == "popular":
            page_url = "shows/popular"
        elif trakt_list == "anticipated":
            page_url = "shows/anticipated"
        elif trakt_list == "collected":
            page_url = "shows/collected"
        elif trakt_list == "watched":
            page_url = "shows/watched"
        elif trakt_list == "played":
            page_url = "shows/played"
        elif trakt_list == "recommended":
            page_url = "recommendations/shows"
        elif trakt_list == "newshow":
            page_url = "calendars/all/shows/new/{0}/30".format(datetime.date.today().strftime("%Y-%m-%d"))
        elif trakt_list == "newseason":
            page_url = "calendars/all/shows/premieres/{0}/30".format(datetime.date.today().strftime("%Y-%m-%d"))
        else:
            page_url = "shows/anticipated"

        trending_shows = []
        black_list = False
        try:
            trending_shows, black_list = trakt_trending.fetch_trending_shows(trakt_list, page_url)
        except Exception as error:
            logger.warning(f"Could not get trending shows: {error}")

        return t.render(black_list=black_list, trending_shows=trending_shows)

    def getTrendingShowImage(self):
        indexer_id = self.get_body_argument("indexerId")
        image_url = sickchill.indexer.series_poster_url_by_id(indexer_id)
        if image_url:
            image_path = trakt_trending.get_image_path(trakt_trending.get_image_name(indexer_id))
            trakt_trending.cache_image(image_url, image_path)
            return indexer_id

    def popularShows(self):
        """
        Fetches data from IMDB to show a list of popular shows.
        """
        t = PageTemplate(rh=self, filename="addShows_popularShows.mako")
        try:
            popular_shows = imdb_popular.fetch_popular_shows()

            popular_shows = {
                show
                for show in popular_shows
                if show.getID() and show.getID().strip("tt") not in {show.imdb_id.strip("tt") for show in settings.show_list if show.imdb_id}
            }
            for show in popular_shows:
                show.setdefault("rating", "0.0")
                show.setdefault("votes", "0")

            imdb_exception = None
        except Exception as error:
            logger.warning(f"Could not get popular shows: {error}")
            logger.debug(traceback.format_exc())
            popular_shows = None
            imdb_exception = error

        return t.render(
            title=_("Popular Shows"),
            header=_("Popular Shows"),
            popular_shows=popular_shows,
            imdb_exception=imdb_exception,
            imdb_url=imdb_popular.imdb_url,
            topmenu="home",
            controller="addShows",
            action="popularShows",
        )

    def favoriteShows(self):
        """
        Fetches data from IMDB to show a list of popular shows.
        """
        t = PageTemplate(rh=self, filename="addShows_favoriteShows.mako")
        error = None

        if self.get_body_argument("submit", None):
            tvdb_user = self.get_body_argument("tvdb_user")
            tvdb_user_key = filters.unhide(settings.TVDB_USER_KEY, self.get_body_argument("tvdb_user_key"))
            if tvdb_user and tvdb_user_key:
                if tvdb_user != settings.TVDB_USER or tvdb_user_key != settings.TVDB_USER_KEY:
                    favorites.test_user_key(tvdb_user, tvdb_user_key, 1)

        try:
            favorite_shows = favorites.fetch_indexer_favorites()
        except Exception as error:
            logger.exception(traceback.format_exc())
            logger.warning(_("Could not get favorite shows: {error}").format(error=error))
            favorite_shows = None

        return t.render(
            title=_("Favorite Shows"),
            header=_("Favorite Shows"),
            favorite_shows=favorite_shows,
            favorites_exception=error,
            topmenu="home",
            controller="addShows",
            action="popularShows",
        )

    def addShowToBlacklist(self):
        # URL parameters

        indexer_id = self.get_query_argument("indexer_id")
        if not indexer_id:
            raise HTTPError(404)

        data = {"shows": [{"ids": {"tvdb": indexer_id}}]}

        trakt_api = TraktAPI(settings.SSL_VERIFY, settings.TRAKT_TIMEOUT)

        trakt_api.traktRequest("users/" + settings.TRAKT_USERNAME + "/lists/" + settings.TRAKT_BLACKLIST_NAME + "/items", data, method="POST")

        return self.redirect("/addShows/trendingShows/")

    def existingShows(self):
        """
        Prints out the page to add existing shows from a root dir
        """
        t = PageTemplate(rh=self, filename="addShows_addExistingShow.mako")
        return t.render(
            enable_anime_options=False, title=_("Existing Show"), header=_("Existing Show"), topmenu="home", controller="addShows", action="addExistingShow"
        )

    def addShowByID(self):
        indexer_id = self.get_query_argument("indexer_id")
        show_name = self.get_query_argument("show_name")
        indexer = self.get_query_argument("indexer", default="TVDB")

        def add_error(in_list: TVShow = None) -> None:
            title = f"Unable to add {show_name}"

            message = f"Could not add {show_name} with {indexer}:{indexer_id}. We were unable to locate the tvdb id at this time."
            if in_list:
                message = f"{in_list.name} with {in_list.indexerid} is already in your show list."

            logger.info(" ".join([title, message]))
            ui.notifications.error(title, message)

            return self.redirect("/home/")

        if indexer != "TVDB":
            indexer_id = helpers.tvdbid_from_remote_id(indexer_id, indexer.upper())
            if not indexer_id:
                return add_error()

        existing = Show.find(settings.show_list, indexer_id)
        if try_int(indexer_id) <= 0 or existing:
            return add_error(existing)

        return self.newShow("|".join([str(1), "", indexer_id, ""]), [], search_string=show_name)

    def addNewShow(
        self,
        rootDir=None,
        defaultStatus=None,
        quality_preset=None,
        anyQualities=None,
        bestQualities=None,
        season_folders=None,
        subtitles=None,
        subtitles_sc_metadata=None,
        other_shows=None,
        skipShow=None,
        anime=None,
        scene=None,
        blacklist=None,
        whitelist=None,
        defaultStatusAfter=None,
        **_kwargs,
    ):
        """
        Receive tvdb id, dir, and other options and create a show from them. If extra show dirs are
        provided then it forwards back to newShow, if not it goes to /home.
        """

        indexer_language = self.get_body_argument("indexerLang", default=settings.INDEXER_DEFAULT_LANGUAGE)

        # grab our list of other dirs if given
        other_shows = self.get_arguments("other_shows")
        full_show_path = self.get_argument("fullShowPath", default=None)

        def finishAddShow():
            # if there are no extra shows then go home
            if not other_shows:
                return self.redirect("/home/")

            # peel off the next one
            next_show = other_shows[0]
            remaining_shows = other_shows[1:]

            # go to add the next show
            return self.newShow(next_show, remaining_shows)

        # if we're skipping then behave accordingly
        if skipShow:
            return finishAddShow()
        else:
            which_series = self.get_argument("whichSeries")

        # sanity check on our inputs
        if (not rootDir and not full_show_path) or not which_series:
            return _("Missing params, no Indexer ID or folder: {show_to_add} and {root_dir}/{show_path}").format(
                show_to_add=which_series, root_dir=rootDir, show_path=full_show_path
            )

        # figure out what show we're adding and where
        series_pieces = which_series.split("|")
        if (which_series and rootDir) or (which_series and full_show_path and len(series_pieces) > 1):
            if len(series_pieces) < 6:
                logger.error("Unable to add show due to show selection. Not enough arguments: {0}".format((repr(series_pieces))))
                ui.notifications.error(_("Unknown error. Unable to add show due to problem with show selection."))
                return self.redirect("/addShows/existingShows/")

            indexer = int(series_pieces[1])
            indexer_id = int(series_pieces[3])
            # Show name was sent in UTF-8 in the form
            show_name = series_pieces[4]
        else:
            # if no indexer was provided use the default indexer set in General settings
            indexer = int(self.get_argument("providedIndexer", default=settings.INDEXER_DEFAULT))
            indexer_id = int(which_series)
            show_name = os.path.basename(os.path.normpath(full_show_path))

        # use the whole path if it's given, or else append the show name to the root dir to get the full show path
        if full_show_path:
            show_dir = os.path.normpath(full_show_path)
            extra_check_dir = show_dir
        else:
            folder_name = show_name
            s = sickchill.indexer.series_by_id(indexerid=indexer_id, indexer=indexer, language=indexer_language)
            if settings.ADD_SHOWS_WITH_YEAR and s.firstAired:
                try:
                    year = "({0})".format(dateutil.parser.parse(s.firstAired).year)
                    if year not in folder_name:
                        folder_name = "{0} {1}".format(s.seriesName, year)
                except (TypeError, ValueError):
                    logger.info(_("Could not append the show year folder for the show: {0}").format(folder_name))

            show_dir = os.path.join(rootDir, sanitize_filename(folder_name))
            extra_check_dir = os.path.join(rootDir, sanitize_filename(show_name))

        # blanket policy - if the dir exists you should have used "add existing show"
        if (os.path.isdir(show_dir) or os.path.isdir(extra_check_dir)) and not full_show_path:
            ui.notifications.error(_("Unable to add show"), _("Folder {show_dir} exists already").format(show_dir=show_dir))
            return self.redirect("/addShows/existingShows/")

        # don't create show dir if config says not to
        if settings.ADD_SHOWS_WO_DIR:
            logger.info(f"Skipping initial creation of {show_dir} due to config.ini setting")
        else:
            dir_exists = helpers.makeDir(show_dir)
            if not dir_exists:
                logger.exception(f"Unable to create the folder {show_dir}, can't add the show")
                ui.notifications.error(_("Unable to add show"), _("Unable to create the folder {show_dir}, can't add the show").format(show_dir=show_dir))
                # Don't redirect to default page because user wants to see the new show
                return self.redirect("/home/")
            else:
                helpers.chmodAsParent(show_dir)

        # prepare the inputs for passing along
        scene = config.checkbox_to_value(scene)
        anime = config.checkbox_to_value(anime)
        season_folders = config.checkbox_to_value(season_folders)
        subtitles = config.checkbox_to_value(subtitles)
        subtitles_sc_metadata = config.checkbox_to_value(subtitles_sc_metadata)

        if whitelist:
            whitelist = short_group_names(whitelist)
        else:
            whitelist = []
        if blacklist:
            blacklist = short_group_names(blacklist)
        else:
            blacklist = []

        if not anyQualities:
            anyQualities = []
        if not bestQualities or try_int(quality_preset, None):
            bestQualities = []
        if not isinstance(anyQualities, list):
            anyQualities = [anyQualities]
        if not isinstance(bestQualities, list):
            bestQualities = [bestQualities]
        newQuality = Quality.combineQualities([int(q) for q in anyQualities], [int(q) for q in bestQualities])

        # add the show
        settings.showQueueScheduler.action.add_show(
            indexer,
            indexer_id,
            show_dir=show_dir,
            default_status=int(defaultStatus),
            quality=newQuality,
            season_folders=season_folders,
            lang=indexer_language,
            subtitles=subtitles,
            subtitles_sc_metadata=subtitles_sc_metadata,
            anime=anime,
            scene=scene,
            paused=None,
            blacklist=blacklist,
            whitelist=whitelist,
            default_status_after=int(defaultStatusAfter),
            root_dir=rootDir,
        )
        ui.notifications.message(_("Show added"), _("Adding the specified show into {show_dir}").format(show_dir=show_dir))

        return finishAddShow()

    @staticmethod
    def split_extra_show(extra_show):
        if not extra_show:
            return None, None, None, None

        split_vals = extra_show.split("|")
        if len(split_vals) == 1:
            indexer = settings.INDEXER_DEFAULT
            show_dir = split_vals[0]
            return indexer, show_dir, None, None
        elif len(split_vals) < 4:
            indexer = split_vals[0]
            show_dir = split_vals[1]
            return indexer, show_dir, None, None

        indexer = split_vals[0]
        show_dir = split_vals[1]
        indexer_id = split_vals[2]
        show_name = "|".join(split_vals[3:])

        return indexer, show_dir, indexer_id, show_name

    def addExistingShows(self):
        """
        Receives a dir list and add them. Adds the ones with given TVDB IDs first, then forwards
        along to the newShow page.
        """

        # grab a list of other shows to add, if provided
        shows_to_add = self.get_arguments("shows_to_add")

        indexer_id_given = []
        dirs_only = []
        # separate all the ones with Indexer IDs
        for cur_dir in shows_to_add:
            if "|" in cur_dir:
                split_vals = cur_dir.split("|")
                if len(split_vals) < 3:
                    dirs_only.append(cur_dir)
            if "|" not in cur_dir:
                dirs_only.append(cur_dir)
            else:
                indexer, show_dir, indexer_id, show_name = self.split_extra_show(cur_dir)

                if not show_dir or not indexer_id or not show_name:
                    continue

                indexer_id_given.append((int(indexer), show_dir, int(indexer_id), show_name))

        # if they want me to prompt for settings then I will just carry on to the newShow page
        if shows_to_add and config.checkbox_to_value(self.get_argument("promptForSettings")):
            return self.newShow(shows_to_add[0], shows_to_add[1:])

        # if they don't want me to prompt for settings then I can just add all the nfo shows now
        num_added = 0
        for cur_show in indexer_id_given:
            indexer, show_dir, indexer_id, show_name = cur_show

            if indexer is not None and indexer_id is not None:
                # add the show
                settings.showQueueScheduler.action.add_show(
                    indexer,
                    indexer_id,
                    show_dir,
                    default_status=settings.STATUS_DEFAULT,
                    quality=settings.QUALITY_DEFAULT,
                    season_folders=settings.SEASON_FOLDERS_DEFAULT,
                    subtitles=settings.SUBTITLES_DEFAULT,
                    anime=settings.ANIME_DEFAULT,
                    scene=settings.SCENE_DEFAULT,
                    default_status_after=settings.STATUS_DEFAULT_AFTER,
                )
                num_added += 1

        if num_added:
            ui.notifications.message(_("Shows Added"), _("Automatically added {num_shows} from their existing metadata files").format(num_shows=str(num_added)))

        # if we're done then go home
        if not dirs_only:
            return self.redirect("/home/")

        # for the remaining shows we need to prompt for each one, so forward this on to the newShow page
        return self.newShow(dirs_only[0], dirs_only[1:])



from typing import TYPE_CHECKING
from urllib.parse import urlparse

import qbittorrentapi

from sickchill import logger, settings
from sickchill.oldbeard.clients.generic import GenericClient

if TYPE_CHECKING:  # pragma: no cover
    from sickchill.oldbeard.classes import TorrentSearchResult


class Client(GenericClient):
    def __init__(self, host: str = None, username: str = None, password: str = None):
        super().__init__("qBittorrent", host, username, password)
        parsed_url = urlparse(self.host or settings.TORRENT_HOST)
        self.host = parsed_url.hostname
        self.port = parsed_url.port
        self.api = qbittorrentapi.Client(
            host=self.host,
            port=self.port or None,
            username=self.username or settings.TORRENT_USERNAME,
            password=self.password or settings.TORRENT_PASSWORD,
            VERIFY_WEBUI_CERTIFICATE=settings.TORRENT_VERIFY_CERT,
        )

    def _get_auth(self):
        try:
            if not self.api.is_logged_in:
                self.api.auth_log_in()
        except (qbittorrentapi.LoginFailed, qbittorrentapi.APIConnectionError):
            return False
        return True

    def test_client_connection(self):
        try:
            if not self.api.is_logged_in:
                self.api.auth_log_in()
        except (qbittorrentapi.LoginFailed, qbittorrentapi.APIConnectionError) as error:
            return False, f"Failed to authenticate with {self.name}, {error}"
        return True, "Success: Connected and Authenticated"

    @staticmethod
    def __torrent_args(result: "TorrentSearchResult") -> dict:
        ratio_limit_float = max(float(result.ratio or 0), 0)

        return dict(
            save_path=settings.TORRENT_PATH or None,
            download_path=settings.TORRENT_PATH_INCOMPLETE or None,
            category=(settings.TORRENT_LABEL, settings.TORRENT_LABEL_ANIME)[result.show.is_anime] or settings.TORRENT_LABEL,
            is_paused=settings.TORRENT_PAUSED,
            ratio_limit=(None, ratio_limit_float)[ratio_limit_float > 0],
            seeding_time_limit=(None, 60 * int(settings.TORRENT_SEED_TIME))[int(settings.TORRENT_SEED_TIME) > 0],
            tags=("sickchill", "sickchill-anime")[result.show.is_anime],
        )

    def __add_trackers(self, result: "TorrentSearchResult"):
        if settings.TRACKERS_LIST and result.provider.public:
            trackers = list({x.strip() for x in settings.TRACKERS_LIST.split(",") if x.strip()})
            if trackers:
                logger.debug(f"Adding trackers to public torrent")
                return self.api.torrents_add_trackers(torrent_hash=result.hash.lower(), urls=trackers)

    def _add_torrent_uri(self, result: "TorrentSearchResult"):
        logger.debug(f"Posting as url with {self.__torrent_args(result)}")
        action = self.api.torrents_add(urls=[result.url], **self.__torrent_args(result))
        self.__add_trackers(result)
        return action

    def _add_torrent_file(self, result: "TorrentSearchResult"):
        logger.debug(f"Posting as file with {self.__torrent_args(result)}")
        action = self.api.torrents_add(torrent_files=[result.content], **self.__torrent_args(result))
        self.__add_trackers(result)
        return action

    def _set_torrent_priority(self, result: "TorrentSearchResult"):
        if not self.api.app_preferences().get("queueing_enabled"):
            return True
        return (self.api.torrents_decrease_priority, self.api.torrents_increase_priority)[result.priority == 1](result.hash.lower())



import logging

import pytest
from packaging import version as packaging_version

from sickchill import settings
from sickchill.update_manager import pip


@pytest.fixture()
def updater(caplog):
    with caplog.at_level(logging.WARNING, logger="root"):
        caplog.set_level(logging.DEBUG, logger="sickchill")
        caplog.set_level(logging.WARNING, logger="cacheyou")
        caplog.set_level(logging.WARNING, logger="urllib3")
        fixture = pip.PipUpdateManager()
        fixture.version_text = "2022.9.14"
        yield fixture


class TestPipUpdateManager:
    def test_get_current_version(self, updater):
        assert updater.get_current_version() == packaging_version.parse(updater.version_text)

    def test_get_clean_version(self, updater):
        assert updater.get_clean_version() == updater.version_text
        assert updater.get_clean_version(use_version=packaging_version.parse("2022.8.30")) == "2022.8.30"

    def test_get_newest_version(self, updater):
        assert updater.get_newest_version() is not None

    def test_get_version_delta(self, updater):
        assert updater.get_current_version() == packaging_version.parse("2022.9.14")

        assert updater.get_version_delta().startswith("Major:")

        updater.version_text = updater.get_clean_version(use_version=updater.get_newest_version())
        assert updater.get_version_delta() == 0

    def test_set_newest_text(self, updater):
        assert updater.need_update()
        updater.set_newest_text()
        assert settings.SITE_MESSAGES

    def test_need_update(self, updater):
        assert updater.need_update()

        updater.version_text = updater.get_clean_version(use_version=updater.get_newest_version())
        assert not updater.need_update()

    def test_update(self, updater):
        assert updater.update()

    def test_pip_install(self, updater):
        assert updater.pip_install("sickchill")



import re
from urllib.parse import quote_plus

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):
        super().__init__("Zamunda")

        self.username = None
        self.password = None
        self.minseed = 0
        self.minleech = 0

        self.custom_url = None

        self.urls = {
            "base_url": "https://zamunda.net",
            "login": "https://zamunda.net/takelogin.php",
            "search": "https://zamunda.net/bananas?c7=1&c33=1&c55=1&search=%s",
            "change_language": "https://zamunda.net/langchange.php?lang=en",
        }

        self.url = self.urls["base_url"]
        self.categories = "&c7=1&c33=1&c55=1"

        self.cache = tvcache.TVCache(self, min_time=30)  # only poll Zamunda every 30 minutes max

    def _check_auth(self):
        if not self.username or not self.password:
            raise AuthException("Invalid username or password. Check your settings")

        return True

    def login(self):
        if not self.check_and_update_urls():
            return False

        cookie_dict = dict_from_cookiejar(self.session.cookies)
        if cookie_dict.get("uid") and cookie_dict.get("pass"):
            return True

        change_language_response = self.get_url(self.urls["change_language"], returns="text")
        if not change_language_response:
            logger.warning(_("Unable to connect to provider"))
            return False

        login_params = {"username": self.username, "password": self.password}

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning(_("Unable to connect to provider"))
            return False

        if re.search("Login failed!", response):
            logger.warning(_("Invalid username or password. Check your settings"))
            return False

        return True

    def search(self, search_strings):
        results = []

        if not self.login():
            return results

        if "RSS" in search_strings:
            del search_strings["RSS"]

        for mode in search_strings:
            items = []

            logger.debug(_("Search Mode: {mode}").format(mode=mode))
            for search_string in {*search_strings[mode]}:
                search_url = self.urls["search"] % quote_plus(search_string) + self.categories
                logger.debug(_("Search String: {search_string}").format(search_string=search_string))

                data = self.get_url(search_url, returns="text")
                if not data or "Flood protection is activated. Please, wait few seconds to continue." in data:
                    logger.debug(_("Flood protection is activated"))
                    continue

                if data.find("Sorry, nothing found") != -1:
                    logger.debug(_("Data returned from provider does not contain any torrents"))
                    continue

                # Search result page contains some invalid html that prevents html parser from returning all data.
                # We cut everything before the table that contains the data we are interested in thus eliminating
                # the invalid html portions
                try:
                    index = data.lower().index('<table align="center" id="zbtable"')
                except ValueError:
                    logger.debug(_("Could not find table of torrents zbtable"))
                    continue

                data = data[index:]

                with BS4Parser(data) as html:
                    if not html:
                        logger.debug(_("No html data parsed from provider"))
                        continue

                    torrent_rows = []
                    torrent_table = html.find("table", class_="test bottom responsivetable responsetop td_clear roundrss")
                    if torrent_table:
                        torrent_rows = torrent_table("tr")

                    if not torrent_rows:
                        logger.debug(_("Could not find results in returned data"))
                        continue

                    # Skip column headers
                    for result in torrent_rows[1:]:
                        try:
                            cells = result.find_all("td")

                            imdrating = cells[1].find("div", class_="imdrating")
                            download_url = imdrating.select_one("a")["href"]

                            dString = str(download_url)
                            start = dString.rindex("/") + 1
                            end = dString.index(".torrent")
                            title = dString[start:end]

                            torrent_size = cells[5].get_text(strip=True)
                            seeders = try_int(cells[7].get_text(strip=True))
                            leechers = try_int(cells[8].get_text(strip=True))

                            size = convert_size(torrent_size) or -1
                        except (AttributeError, TypeError, KeyError, ValueError, IndexError):
                            continue

                        if not all([title, download_url]):
                            continue

                        # Filter unseeded torrent
                        if seeders < self.minseed or leechers < self.minleech:
                            if mode != "RSS":
                                logger.debug(
                                    _("Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})").format(
                                        title, seeders, leechers
                                    )
                                )
                            continue

                        item = {"title": title, "link": self.url + download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}

                        items.append(item)

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



from datetime import date
from typing import TYPE_CHECKING, Union

from sickchill import settings
from sickchill.helper.exceptions import CantRefreshShowException, CantRemoveShowException, CantUpdateShowException, MultipleShowObjectsException
from sickchill.oldbeard.common import Quality, SKIPPED, WANTED
from sickchill.oldbeard.db import DBConnection

if TYPE_CHECKING:
    from sickchill.tv import TVShow


class Show(object):
    @staticmethod
    def find(shows: list, indexer_id: Union[int, str]) -> Union["TVShow", None]:
        """
        Find a show by its indexer id in the provided list of shows
        :param shows: The list of shows to search in
        :param indexer_id: The indexer id of the desired show
        :return: The desired show if found, ``None`` if not found
        :throw: ``MultipleShowObjectsException`` if multiple shows match the provided ``indexer_id``
        """
        if not (indexer_id and shows):
            return None

        if isinstance(indexer_id, list):
            if not isinstance(indexer_id[0], (int, str)):
                return None

            indexer_ids = [int(x) for x in indexer_id]
        else:
            if not isinstance(indexer_id, (int, str)):
                return None
            indexer_ids = [int(indexer_id)]

        results = [show for show in shows if show.indexerid in indexer_ids]

        if not results:
            return None

        if len(results) == 1:
            return results[0]

        raise MultipleShowObjectsException()

    @staticmethod
    def find_name(shows: list, name: str) -> Union["TVShow", None]:
        """
        Find a show by its namer in the provided list of shows
        :param shows: The list of shows to search in
        :param name: The known name of the desired show
        :return: The desired show if found, ``None`` if not found
        :throw: ``MultipleShowObjectsException`` if multiple shows match the provided ``indexer_id``
        """
        if not (name and shows):
            return None

        if isinstance(name, list):
            names = name
            for item in names:
                if not isinstance(item, str):
                    return None
        else:
            if not isinstance(name, str):
                return None

            names = [name]

        results = [show for show in shows if show.name in names]

        if not results:
            return None

        if len(results) == 1:
            return results[0]

        raise MultipleShowObjectsException()

    @staticmethod
    def overall_stats() -> dict:
        db = DBConnection()
        shows = settings.show_list
        today = date.today().toordinal()

        downloaded_status = Quality.DOWNLOADED + Quality.ARCHIVED
        snatched_status = Quality.SNATCHED + Quality.SNATCHED_PROPER + Quality.SNATCHED_BEST
        total_status = [SKIPPED, WANTED]

        results = db.select("SELECT airdate, status FROM tv_episodes WHERE season > 0 AND episode > 0 AND airdate > 1")

        stats = {
            "episodes": {
                "downloaded": 0,
                "snatched": 0,
                "total": 0,
            },
            "shows": {
                "active": len([show for show in shows if show.paused == 0 and show.status == "Continuing"]),
                "total": len(shows),
            },
        }

        for result in results:
            if result["status"] in downloaded_status:
                stats["episodes"]["downloaded"] += 1
                stats["episodes"]["total"] += 1
            elif result["status"] in snatched_status:
                stats["episodes"]["snatched"] += 1
                stats["episodes"]["total"] += 1
            elif result["airdate"] <= today and result["status"] in total_status:
                stats["episodes"]["total"] += 1

        return stats

    @staticmethod
    def validate_indexer_id(show_or_id: Union["TVShow", str, int], show_list: list = None) -> (Union[str, None], Union["TVShow", None]):
        """
        Check that the provided indexer_id is valid and corresponds with a known show
        :param show_or_id: The indexer id or object to check
        :param show_list: The list of shows to check against
        :return: A tuple containing:
         - an error message if the indexer id is not correct, ``None`` otherwise
         - the show object corresponding to ``indexer_id`` if it exists, ``None`` otherwise
        """

        # NOTE: cannot import TVShow here because of circular import, check this way
        if hasattr(show_or_id, "custom_name"):
            return None, show_or_id

        try:
            indexer_id = int(show_or_id)
        except (TypeError, ValueError):
            return _("Invalid show ID") + f" {show_or_id}", None

        if not show_list:
            show_list = settings.show_list

        try:
            show = Show.find(show_list, indexer_id)
        except MultipleShowObjectsException:
            return "Unable to find the specified show", None

        return None, show

    @staticmethod
    def pause(show_or_id: Union[int, str, "TVShow"], pause: Union[bool, None] = None) -> (Union[str, None], Union["TVShow", None]):
        """
        Change the pause state of a show
        :param show_or_id: The unique id or object of the show to update
        :param pause: ``True`` to pause the show, ``False`` to resume the show, ``None`` to toggle the pause state
        :return: A tuple containing:
         - an error message if the pause state could not be changed, ``None`` otherwise
         - the show object that was updated, if it exists, ``None`` otherwise
        """

        error, show = Show.validate_indexer_id(show_or_id)
        if error is not None:
            return error, show

        if pause is None:
            show.paused = not show.paused
        else:
            show.paused = pause

        show.save_to_db()

        return None, show

    @staticmethod
    def refresh(show_or_id: Union[int, str, "TVShow"], force: bool = False) -> (Union[str, None], Union["TVShow", None]):
        """
        Try to refresh a show
        :param force: Force refresh
        :param show_or_id: The unique id or object of the show to refresh
        :return: A tuple containing:
         - an error message if the show could not be refreshed, ``None`` otherwise
         - the show object that was refreshed, if it exists, ``None`` otherwise
        """

        error, show = Show.validate_indexer_id(show_or_id)
        if error is not None:
            return error, show

        try:
            settings.showQueueScheduler.action.refresh_show(show, force)
        except CantRefreshShowException as exception:
            return str(exception), show

        return None, show

    @staticmethod
    def update(show_or_id: Union[int, str, "TVShow"], force: bool = False) -> (Union[str, None], Union["TVShow", None]):
        """
        Try to delete a show
        :param show_or_id: The unique id or object of the show to delete
        :param force: Force update
        :return: A tuple containing:
         - an error message if the show could not be deleted, ``None`` otherwise
         - the show object that was deleted, if it exists, ``None`` otherwise
        """
        error, show = Show.validate_indexer_id(show_or_id)
        if error is not None:
            return error, show

        try:
            settings.showQueueScheduler.action.update_show(show, bool(force))
        except CantUpdateShowException as exception:
            return str(exception), show

        return None, show

    @staticmethod
    def delete(show_or_id: Union[int, str, "TVShow"], remove_files: bool = False) -> (Union[str, None], Union["TVShow", None]):
        """
        Try to delete a show
        :param show_or_id: The unique id or object of the show to delete
        :param remove_files: ``True`` to remove the files associated with the show, ``False`` otherwise
        :return: A tuple containing:
         - an error message if the show could not be deleted, ``None`` otherwise
         - the show object that was deleted, if it exists, ``None`` otherwise
        """
        error, show = Show.validate_indexer_id(show_or_id)
        if error is not None:
            return error, show

        try:
            settings.showQueueScheduler.action.remove_show(show, bool(remove_files))
        except CantRemoveShowException as exception:
            return str(exception), show

        return None, show



"""
Test XEM
"""

import unittest

import sickchill.oldbeard.db
from sickchill import settings
from sickchill.tv import TVShow
from tests import conftest


class XEMBasicTests(conftest.SickChillTestDBCase):
    """
    Perform basic xem tests
    """

    @staticmethod
    def load_shows_from_db():
        """
        Populates the show_list with shows from the database
        """

        test_main_db_con = sickchill.oldbeard.db.DBConnection()
        sql_results = test_main_db_con.select("SELECT * FROM tv_shows")

        for sql_show in sql_results:
            # noinspection PyBroadException
            try:
                cur_show = TVShow(int(sql_show["indexer"]), int(sql_show["indexer_id"]))
                settings.show_list.append(cur_show)
            except Exception:
                pass

    @staticmethod
    def load_from_db():
        """
        Populates the show_list with shows from the database
        """
        test_main_db_con = sickchill.oldbeard.db.DBConnection()
        sql_results = test_main_db_con.select("SELECT * FROM tv_shows")

        for sql_show in sql_results:
            try:
                cur_show = TVShow(int(sql_show["indexer"]), int(sql_show["indexer_id"]))
                settings.show_list.append(cur_show)
            except Exception as error:
                print(f"There was an error creating the show {error}")


if __name__ == "__main__":
    print("==================")
    print("STARTING - XEM Scene Numbering TESTS")
    print("==================")
    print("######################################################################")

    SUITE = unittest.TestLoader().loadTestsFromTestCase(XEMBasicTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



from sickchill import settings


def hide(value):
    return "hidden_value" if value else ""


def unhide(old, new):
    return (new, old)[new == "hidden_value"] or None


def hidden(condition: bool) -> str:
    if condition:
        return ""
    return "hidden"


def disabled(condition: bool) -> str:
    if condition:
        return ""
    return "disabled"


def checked(condition: bool) -> str:
    if condition:
        return "checked"
    return ""


def selected(condition: bool) -> str:
    if condition:
        return "selected"
    return ""


def filter_shows_being_removed(show_list):
    return {
        show
        for show in show_list
        if not (settings.showQueueScheduler.action.is_in_remove_queue(show) or settings.showQueueScheduler.action.is_being_removed(show))
    }



import os
import subprocess
import sys
from pathlib import Path
from typing import List, Union

from packaging import version as packaging_version
from packaging.version import Version

from sickchill import logger, settings
from sickchill.init_helpers import get_current_version, sickchill_dir
from sickchill.oldbeard import helpers, notifiers

from .abstract import UpdateManagerBase


class PipUpdateManager(UpdateManagerBase):
    def __init__(self):
        self.version_text = get_current_version()
        self.newest_version_text = get_current_version()

        self._newest_version: packaging_version.Version = None
        self.session = helpers.make_session()

    def get_current_version(self) -> Version:
        return packaging_version.parse(self.version_text)

    def get_clean_version(self, use_version: packaging_version.Version = None):
        return str(use_version or self.get_current_version())

    def get_newest_version(self) -> packaging_version.Version:
        self._newest_version = packaging_version.parse(self.session.get("https://pypi.org/pypi/sickchill/json").json()["info"]["version"])
        return self._newest_version

    def get_version_delta(self):
        if not self.need_update():
            return 0

        newest, current = self._newest_version, self.get_current_version()
        return f"Major: {newest.major - current.major}, Minor: {newest.minor - current.minor}, Micro: {newest.micro - current.micro}"

    def set_newest_text(self):
        if self.need_update():
            if self.get_newest_version():
                current = self.get_clean_version()
                newest = self.get_clean_version(self.get_newest_version())
                url = f"https://github.com/{settings.GIT_ORG}/{settings.GIT_REPO}/compare/{current}...{newest}"
            else:
                url = f"https://github.com/{settings.GIT_ORG}/{settings.GIT_REPO}/commits/"

            newest_tag = "newer_version_available"
            update_url = self.get_update_url()
            newest_text = _(
                'There is a <a href="{url}" target="_blank" rel="noreferrer">newer version available</a> &mdash; <a href="{update_url}">Update Now</a>'
            ).format(url=url, update_url=update_url)

        else:
            return

        helpers.add_site_message(newest_text, tag=newest_tag, level="success")

    def need_update(self):
        return self.get_newest_version() > self.get_current_version()

    def update(self):
        logger.info("Updating using pip in your current environment")
        if not self.pip_install("sickchill"):
            return False

        notifiers.notify_update(f"{self._newest_version}")
        return True

    def pip_install(self, packages: Union[List[str], str, Path]) -> bool:
        def subprocess_call(cmd_list):
            try:
                process = subprocess.Popen(
                    cmd_list, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, cwd=os.getcwd()
                )
                stdout, stderr = process.communicate()
                process.wait()
                if stdout or stderr:
                    for line in (stdout or stderr).splitlines():
                        logger.info(line)
            except Exception as error:
                logger.info(f"Unable to run command: {error}")
                return 126
            return process.returncode

        def get_os_id():
            os_release = Path("/etc/os-release").resolve()
            if os_release.is_file():
                from configparser import ConfigParser

                parser = ConfigParser()
                parser.read_string("[DEFAULT]\n" + os_release.read_text())
                try:
                    return parser["DEFAULT"]["ID"]
                except (KeyError, IndexError):
                    pass

        cmd = [
            sys.executable,
            "-m",
            "pip",
            "install",
            "--no-input",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--no-color",
            # "--trusted-host=pypi.org",
            # "--trusted-host=files.pythonhosted.org",
            "-U",
        ]

        os_id = get_os_id()
        if os_id in ("alpine", "ubuntu"):
            cmd.append(f"--find-links=https://wheel-index.linuxserver.io/{os_id}/")

        if os_id == "alpine":
            cmd.append(f"--extra-index-url=https://alpine-wheels.github.io/index")

        elif os_id in ("raspbian", "osmc"):
            cmd.append(f"--extra-index-url=https://www.piwheels.org/simple")

        syno_wheelhouse = sickchill_dir.with_name("wheelhouse")
        if syno_wheelhouse.is_dir():
            logger.debug(f"Found wheelhouse dir at {syno_wheelhouse}")
            cmd.append(f"-f{syno_wheelhouse}")

        if isinstance(packages, list):
            cmd += packages
        elif isinstance(packages, str):
            cmd.append(packages)
        elif isinstance(packages, Path):
            cmd.append(str(packages.resolve()))

        logger.debug(f"pip args: {' '.join(cmd)}")

        result = subprocess_call(cmd)
        if result != 0:  # Not Ok
            logger.info("Trying user site-packages")
            result = subprocess_call(cmd + ["--user"])
            if result != 0:  # Not Ok
                return False
        return True
